/* srpres.c
 *
 * Copyright 2001, 2002 Sun Microsystems, Inc.,
 * Copyright 2001, 2002 BAUM Retec, A.G.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "srpres.h"
#include "SRObject.h"
#include "glib.h"
#include <string.h>
/*#include "srmain.h"*/
#include "srspc.h"
#include "libsrconf.h"
#include "srintl.h"

typedef enum
{
    SRC_PRESENTATION_NAME,
    SRC_PRESENTATION_ROLE,
    SRC_PRESENTATION_ACCELERATOR,
    SRC_PRESENTATION_SHORTCUT,
    SRC_PRESENTATION_LEVEL,
    SRC_PRESENTATION_POSITION_TOP,
    SRC_PRESENTATION_POSITION_LEFT,
    SRC_PRESENTATION_POSITION_RIGHT,
    SRC_PRESENTATION_POSITION_BOTTOM,
    SRC_PRESENTATION_POSITION_WIDTH,
    SRC_PRESENTATION_POSITION_HEIGHT,
    SRC_PRESENTATION_TEXT_CHAR,
    SRC_PRESENTATION_TEXT_BEST,
    SRC_PRESENTATION_TEXT_WORD,
    SRC_PRESENTATION_TEXT_LINE,
    SRC_PRESENTATION_TEXT_SENTENCE,
    SRC_PRESENTATION_TEXT_PARAGRAPH,
    SRC_PRESENTATION_TEXT_DOCUMENT,
    SRC_PRESENTATION_TEXT_SELECTION,
    SRC_PRESENTATION_TEXT_SELECTIONS,
    SRC_PRESENTATION_TEXT_DIFFERENCE,
    SRC_PRESENTATION_STATE_CHECKED,
    SRC_PRESENTATION_STATE_UNCHECKED,
    SRC_PRESENTATION_STATE_ENABLED,
    SRC_PRESENTATION_STATE_EXPANDED,
    SRC_PRESENTATION_STATE_COLLAPSED,
    SRC_PRESENTATION_STATE_MINIMIZED,
    SRC_PRESENTATION_CHILDREN_COUNT,
    SRC_PRESENTATION_VALUE_CRT,
    SRC_PRESENTATION_NONE
}SRCPresentationItemType;


typedef enum
{
    SRC_PRESENTATION_INFO_TYPE_INFO,
    SRC_PRESENTATION_INFO_TYPE_TEXT,
    SRC_PRESENTATION_INFO_TYPE_COND,
    SRC_PRESENTATION_INFO_TYPE_PARANTHESIS,
    SRC_PRESENTATION_INFO_TYPE_COND1,
    SRC_PRESENTATION_INFO_TYPE_SPECIAL,
    SRC_PRESENTATION_INFO_TYPE_CHILDREN,
    SRC_PRESENTATION_INFO_TYPE_NONE
}SRCPresentationInfoType;


typedef struct _SRCPresentationItem
{
    SRCPresentationInfoType	type;
    SRCPresentationItemType	info;
    gpointer			data1;
    gpointer			data2;
    gpointer			voice;
}SRCPresentationItem;

static SRCPresentationItem*
src_presentation_item_new ()
{
    SRCPresentationItem *item;
    
    item = g_new (SRCPresentationItem, 1);
    if (item)
    {
	item->type  = SRC_PRESENTATION_INFO_TYPE_NONE;
	item->info  = SRC_PRESENTATION_NONE;
	item->data1 = item->data2 = NULL;
	item->voice = NULL;    
    }
    return item;
}

static void
src_presentation_item_terminate (SRCPresentationItem *item)
{
    sru_assert (item);
    
    switch (item->type)
    {
	case SRC_PRESENTATION_INFO_TYPE_TEXT:
	    g_free (item->data1);
	    g_free (item->voice);
	    break;
	case SRC_PRESENTATION_INFO_TYPE_COND:
	case SRC_PRESENTATION_INFO_TYPE_COND1:
	    g_free (item->data1);
	    g_free (item->data2);
	    g_free (item->voice);
	    break;
	case SRC_PRESENTATION_INFO_TYPE_CHILDREN:
	    if (item->data1)
	    {
		SRCPresentationItem **item1;
		gint i;
		item1 = (SRCPresentationItem **)item->data1;
		for (i = 0; item1[i]; i++)	
		    src_presentation_item_terminate (item1[i]);
		g_free (item->data1);
	    }
	    g_free (item->data2);
	    break;
	case SRC_PRESENTATION_INFO_TYPE_INFO:
	case SRC_PRESENTATION_INFO_TYPE_PARANTHESIS:    
	    g_free (item->voice);
	    break;
	case SRC_PRESENTATION_INFO_TYPE_SPECIAL:
	    g_free (item->data1);
	    break;
	default:
	    sru_assert_not_reached ();
	    break;
    }
    g_free (item);    
}


static void
src_presentation_items_terminate (SRCPresentationItem **items)
{
    gint i;
    sru_assert (items);
    
    for (i = 0; items[i]; i++)
    {
        src_presentation_item_terminate (items[i]);
    }

    g_free (items);
}




static gboolean
src_presentation_write_formated_string_to_gconf (const gchar *device,
						 const gchar *role,
						 const gchar *event,
						 const gchar *string)
{
    gboolean rv;
    gchar *key;
    
    sru_assert (device && role && event && string);
    
    key = g_strconcat (device, "_", role, "_", event, NULL);
    rv  = srconf_set_data(key, CFGT_STRING, (gchar *) string, PRESENTATION_PATH);
    g_free (key);
    
    return rv;
}



static gchar*
src_presentation_read_formated_string_from_gconf (const gchar *device,
						  const gchar *role,
						  const gchar *event)
{
    gchar *key = NULL;
    gchar *rv  = NULL;
    
    sru_assert (device && role && event);
    
    key = g_strconcat (device, "_", role, "_", event, NULL);
    srconf_get_data_with_default (key, CFGT_STRING, &rv, NULL, PRESENTATION_PATH);
    g_free (key);

    return rv;
}


static gchar*
src_presentation_try_read_formated_string_from_gconf (const gchar *device,
						      const gchar *role,
						      const gchar *event)
{
    gchar *rv = NULL;
    
    sru_assert (device && role && event);
    
    rv = src_presentation_read_formated_string_from_gconf (device, role, event);
    if (!rv)
	rv = src_presentation_read_formated_string_from_gconf (device, role, "generic");
    if (!rv)
	rv = src_presentation_read_formated_string_from_gconf (device, "generic", event);
    if (!rv)
	rv = src_presentation_read_formated_string_from_gconf (device, "generic", "generic");

    return rv;    
}


static gchar*
src_presentation_read_formated_string_from_default (const gchar *device,
						    const gchar *role,
						    const gchar *event)
{
    sru_assert (device && role && event);
    
    if (strcmp (device, "braille") == 0)
	return g_strdup ("%role%%name%");
    else if (strcmp (device, "speech") == 0)
	return g_strdup ("%name<voice=name>%%role<voice=role>%");
    else
	return g_strdup ("%object<AlignmentX=\"0\" AlignmentY=\"0\">%");

    return NULL;
}


static gchar*
src_presentation_try_read_formated_string_from_default (const gchar *device,
							const gchar *role,
							const gchar *event)
{
    gchar *rv = NULL;
    
    sru_assert (device && role && event);
    
    rv = src_presentation_read_formated_string_from_default (device, role, event);
    if (!rv)
	rv = src_presentation_read_formated_string_from_default (device, role, "generic");
    if (!rv)
	rv = src_presentation_read_formated_string_from_default (device, "generic", event);
    if (!rv)
	rv = src_presentation_read_formated_string_from_default (device, "generic", "generic");
	
    return rv;    
}


static SRCPresentationItemType
src_presentation_get_item_type (const gchar *item)
{
    static struct
    {
	SRCPresentationItemType type;
	gchar *name;
    }src_presentation_type_name [] = 
	{
	    {SRC_PRESENTATION_NAME,		"name"},
	    {SRC_PRESENTATION_ROLE,		"role"},
	    {SRC_PRESENTATION_ACCELERATOR,	"accelerator"},
	    {SRC_PRESENTATION_SHORTCUT,		"shortcut"},
	    {SRC_PRESENTATION_LEVEL,		"level"},
	    {SRC_PRESENTATION_POSITION_TOP, 	"position:top"},
	    {SRC_PRESENTATION_POSITION_LEFT,	"position:left"},
	    {SRC_PRESENTATION_POSITION_RIGHT,	"position:right"},
	    {SRC_PRESENTATION_POSITION_BOTTOM,	"position:bottom"},
	    {SRC_PRESENTATION_POSITION_WIDTH,	"position:width"},
	    {SRC_PRESENTATION_POSITION_HEIGHT,	"position:height"},
	    {SRC_PRESENTATION_TEXT_CHAR,	"text:char"},
	    {SRC_PRESENTATION_TEXT_BEST,	"text:best"},
	    {SRC_PRESENTATION_TEXT_WORD,	"text:word"},
	    {SRC_PRESENTATION_TEXT_LINE,	"text:line"},
	    {SRC_PRESENTATION_TEXT_SENTENCE,	"text:sentence"},
	    {SRC_PRESENTATION_TEXT_PARAGRAPH,	"text:paragraph"},
	    {SRC_PRESENTATION_TEXT_DOCUMENT,	"text:document"},
	    {SRC_PRESENTATION_TEXT_SELECTION,	"text:selection"},
	    {SRC_PRESENTATION_TEXT_SELECTIONS,	"text:selections"},
	    {SRC_PRESENTATION_TEXT_DIFFERENCE,	"text:difference"},
	    {SRC_PRESENTATION_STATE_CHECKED,	"state:checked"},
	    {SRC_PRESENTATION_STATE_UNCHECKED,	"state:unchecked"},
	    {SRC_PRESENTATION_STATE_ENABLED,	"state:enabled"},
	    {SRC_PRESENTATION_STATE_EXPANDED,	"state:expanded"},
	    {SRC_PRESENTATION_STATE_COLLAPSED,	"state:collapsed"},
	    {SRC_PRESENTATION_STATE_MINIMIZED,	"state:minimized"},
	    {SRC_PRESENTATION_CHILDREN_COUNT,	"childcount"},
	    {SRC_PRESENTATION_VALUE_CRT,	"value:crt"}
	};
    gint i;
    SRCPresentationItemType rv = SRC_PRESENTATION_NONE;

    sru_assert (item);
    
    for (i = 0; i < G_N_ELEMENTS (src_presentation_type_name); i++)
    {
	if (strcmp (src_presentation_type_name[i].name, item) == 0)
	{
	    rv = src_presentation_type_name[i].type;
	}
    }
    
    return rv;
}

static gboolean
src_presentation_check_parameter_ (const gchar *parameter,
				  const gchar *device)
{
    sru_assert (parameter && device);

    /* fprintf (stderr, "\nPARAMETER:%s:", parameter); */

    /* 6 = strlen ("voice=") */    
    if (strncmp (parameter, "voice=", 6) != 0)
	return FALSE;

    if (parameter[6] == '\0')
	return FALSE;

    return TRUE;
}


static gboolean
src_presentation_check_parameter (const gchar *start,
				  const gchar *end,
				  const gchar *device)
{
    gboolean rv = FALSE;
    gchar *parameter;
    
    sru_assert (start && end && device && end > start);
    
    parameter = g_strndup (start, end - start + 1);
    if (parameter)
	rv = src_presentation_check_parameter_ (parameter, device);
    g_free (parameter);
    
    return rv;
}


static gboolean
src_presentation_check_info_ (const gchar *info,
			      const gchar *device)
{
    gchar *start, *end;
    gchar *item;

    sru_assert (info && device);
    
    /* fprintf (stderr, "\nINFO:%s:", info); */
    
    start = strstr (info, "<");
    if (start)
    {
        end = strstr (start, ">");
	if (!end || end[1] != '\0')
	    return FALSE;
	if (!src_presentation_check_parameter (start + 1, end - 1, device))
	    return FALSE;
    }
    else
    {
	start = (gchar *)info + strlen (info);
    }
    item = g_strndup (info, start - info);
    if (!item)
	return FALSE;
    if (src_presentation_get_item_type (item) == SRC_PRESENTATION_NONE)
    {
	g_free (item);
	return FALSE;
    }
    g_free (item);
    
    return TRUE;
}


static gboolean
src_presentation_check_info (const gchar *start,
			     const gchar *end,
			     const gchar *device)
{
    gboolean rv = FALSE;
    gchar *info;
    
    sru_assert (start && end && device && end > start);
    
    info = g_strndup (start, end - start + 1);
    if (info)
	rv = src_presentation_check_info_ (info, device);
    g_free (info);
    
    return rv;
}

static gboolean
src_presentation_check_text_ (const gchar *text, 
			      const gchar *device)
{
    gchar *start, *end;

    sru_assert (text && device);

    /* fprintf (stderr, "\nTEXT:%s:", text); */
    
    start = strstr (text, "<");
    if (start)
    {
	if (start - text == 0)
	    return FALSE;
	end = strstr (start, ">");
	if (!end || end[1] != '\0')
	    return FALSE;    
	if (!src_presentation_check_parameter (start + 1, end - 1, device))
	    return FALSE;
    }
    return TRUE;
}



static gboolean
src_presentation_check_text (const gchar *start,
			     const gchar *end,
			     const gchar *device)
{
    gboolean rv = FALSE;
    gchar *text;
    
    sru_assert (start && end && device && end > start);
    
    text = g_strndup (start, end - start + 1);
    if (text)
	rv = src_presentation_check_text_ (text, device);
    g_free (text);
    
    return rv;
}


static gboolean
src_presentation_check_special_ (const gchar *special,
				 const gchar *device)
{
    gint i, j, len;
    sru_assert (special && device);
    
    /* fprintf (stderr, "\nSPECIAL:%s:", special); */
    len = strlen (special);
    for (i = 0; i < len; i++)
    {
	if (special[i] < '0' || special[i] > '8')
	    return FALSE;
	for (j = i + 1; j < len; j++)
	    if (special[i] == special[j])
		return FALSE;
    }
    return TRUE;
}


static gboolean
src_presentation_check_special (const gchar *start,
				const gchar *end,
				const gchar *device)
{
    gboolean rv = FALSE;
    gchar *special;
    
    sru_assert (start && end && device && end > start);
    
    special = g_strndup (start, end - start + 1);
    if (special)
	rv = src_presentation_check_special_ (special, device);
    g_free (special);
    
    return rv;
}


static gboolean
src_presentation_check_cond_ (const gchar *cond, 
			      const gchar *device)
{
    gchar *start, *end;
    gint len = 0;
    sru_assert (cond && device);

    /* fprintf (stderr, "\nCOND:%s:", cond); */
    
    start = strstr (cond, "?");
    if (start)
    {
	if (!src_presentation_check_info (cond, start - 1, device))
	    return FALSE;
    }
    if (start[1] != '\"')
	return FALSE;
    start++;
    end = strstr (start + 1, "\"");
    if (!end)
	return FALSE;
    len = end - start - 1;
	
    if (end[1] != ':')
	return FALSE;
    if (end[2] != '\"')
	return FALSE;
    start = end + 2;
    end = strstr (start + 1, "\"");
    if (!end)
	return FALSE;

    len = len + end - start - 1;

    if (end[1] == '<')
    {
	start = end + 1;
	end = strstr (start, ">");
	if (!end)
	    return FALSE;
    	if (!src_presentation_check_parameter (start + 1, end - 1, device))
	    return FALSE;
    }
    
    if (end[1] != '\0')
	return FALSE;
    if (len == 0)
	return FALSE;
    return TRUE;
}


static gboolean
src_presentation_check_cond (const gchar *start,
			     const gchar *end,
			     const gchar *device)
{
    gboolean rv = FALSE;
    gchar *cond;
    
    sru_assert (start && end && device && end > start);
    
    cond = g_strndup (start, end - start + 1);
    if (cond)
	rv = src_presentation_check_cond_ (cond, device);
    g_free (cond);
    
    return rv;
}

static gboolean src_presentation_is_formated_string_ok (const gchar *string, const gchar *device);

static gboolean
src_presentation_check_children_ (const gchar *children,
				  const gchar *device)
{
    gchar *start, *end;
    gchar *tmp;

    sru_assert (children && device);

    /* fprintf (stderr, "\nCHILDREN:%s:", children); */

    /* 8 = strlen ("children") */
    if (children[8] != '[')
	return FALSE;
    
    start = (gchar*)children + 8;
    end = strstr (start, "]");
    if (!end)
    	return FALSE;

    tmp = g_strndup (start + 1, end - start - 1);
    if (!src_presentation_is_formated_string_ok (tmp, device))
    {
	g_free (tmp);
	return FALSE;
    }
    g_free (tmp);

    if (end[1] != '[')
	return FALSE;
    start = end + 1;
    end = strstr (start, "]");
    if (!end)
	return FALSE;
    if (end - start > 1)
    {
	tmp = g_strndup (start + 1, end - start - 1);
	if (!src_presentation_is_formated_string_ok (tmp, device))
	{
	    g_free (tmp);
	    return FALSE;
	}
	g_free (tmp);
    }

    if (end[1] != '\0')
	return FALSE;

    return TRUE;
}



static gboolean
src_presentation_check_children (const gchar *start,
				 const gchar *end,
				 const gchar *device)
{
    gboolean rv = FALSE;
    gchar *children;
    
    sru_assert (start && end && device && end > start);
    
    children = g_strndup (start, end - start + 1);
    if (children)
	rv = src_presentation_check_children_ (children, device);
    g_free (children);
    
    return rv;
}



static gboolean
src_presentation_is_formated_string_ok (const gchar *string, 
					const gchar *device)
{
    gchar *crt;
    gchar *next;

    sru_assert (string);

    crt = (gchar*)string;
    while (crt && crt[0])
    {
	if (crt[0] != '%')
	{
	    gchar *tmp;
	    tmp = strstr (crt, "%");
	    if (!tmp)
		tmp = crt + strlen (crt) - 1;
	    if (!src_presentation_check_text (crt, tmp - 1, device))
		return FALSE;
	    crt = tmp;
	    continue;
	}
	
	next = strstr (crt + 1, "%");
	if (!next)
	    return FALSE;
	if (crt[1] == '!')
	{
	    if (next[-1] != '!' || !src_presentation_check_special (crt + 2, next - 2, device))
		return FALSE;
	}
	else if (crt[1] == '?')
	{
	    if (next[-1] != '?' || !src_presentation_check_cond (crt + 2, next - 2, device))
		return FALSE;
	}
	else if (crt[1] == '1')
	{
	    if (next[-1] != '1' || !src_presentation_check_cond (crt + 2, next - 2, device))
		return FALSE;
	}
	else if (crt[1] == '(')
	{
	    if (next[-1] != ')' || !src_presentation_check_info (crt + 2, next - 2, device))
		return FALSE;
	}
	else if (strncmp (crt + 1, "children", 8) == 0)
	{
	    next = strstr (crt, "]");
	    if (next)
		next = strstr (next + 1, "]");
	    if (!next)
		return FALSE;
		
	    next++;
	    if (!src_presentation_check_children (crt + 1, next - 1, device))
		return FALSE;
	}
	else
	{
	    if (!src_presentation_check_info (crt + 1, next - 1, device))
		return FALSE;
	}
	crt = next + 1;
    }
    return TRUE;
}


static SRCPresentationItem*
src_presentation_get_presentation_item_from_text_ (const gchar *text,
						   const gchar *device)
{
    gchar *start, *end;
    SRCPresentationItem *item;

    sru_assert (text && device);
    
    item = src_presentation_item_new ();
    if (!item)
        return NULL;

    item->type  = SRC_PRESENTATION_INFO_TYPE_TEXT;
    start = strstr (text, "<");
    if (start)
    {
	if (strncmp (start + 1, "voice", 5) == 0)
	{
	    end = strstr (start + 6, ">");
	    item->voice = g_strndup (start + 7, strlen (start) - 8);
	}
    }
    else
    {
	start = (gchar *) text + strlen (text);
    }
    item->data1 = g_strndup (text, start - text);
    return item;
}


static SRCPresentationItem*
src_presentation_get_presentation_item_from_text (const gchar *start,
						  const gchar *end,
						  const gchar *device)
{
    SRCPresentationItem *item = NULL;
    gchar *text;
    
    sru_assert (start && end && device && end > start);
    
    text = g_strndup (start, end - start + 1);
    if (text)
	item = src_presentation_get_presentation_item_from_text_ (text, device);
    g_free (text);
    
    return item;
}


static SRCPresentationItem*
src_presentation_get_presentation_item_from_special_ (const gchar *special,
						      const gchar *device)
{
    SRCPresentationItem *item;

    sru_assert (special && device);
    
    item = src_presentation_item_new ();
    if (!item)
        return NULL;
     
    item->type  = SRC_PRESENTATION_INFO_TYPE_SPECIAL;
    item->data1 = g_strdup (special);

    return item;
}



static SRCPresentationItem*
src_presentation_get_presentation_item_from_special (const gchar *start,
						     const gchar *end,
						     const gchar *device)
{
    SRCPresentationItem *item = NULL;
    gchar *special;
    
    sru_assert (start && end && device && end > start);
    
    special = g_strndup (start, end - start + 1);
    if (special)
	item = src_presentation_get_presentation_item_from_special_ (special, device);
    g_free (special);
    
    return item;
}

static SRCPresentationItem*
src_presentation_get_presentation_item_from_cond_ (const gchar *cond,
						   const gchar *device)
{
    gchar *start, *end;
    SRCPresentationItem *item;
    gchar *tmp;

    sru_assert (cond && device);
    
    item = src_presentation_item_new ();
    if (!item)
        return NULL;

    item->type  = SRC_PRESENTATION_INFO_TYPE_COND;
    start = strstr (cond, "?");
    
    tmp = g_strndup (cond, start - cond);
    if (!tmp)
	return NULL;
    item->info = src_presentation_get_item_type (tmp);
    g_free (tmp);
    
    start++;
    end = strstr (start + 1, "\"");
    if (end - start > 1)
    	item->data1 = g_strndup (start + 1, end - start - 1);
    
    start = end + 2;
    end = strstr (start + 1, "\"");
    if (end - start > 1)
    	item->data2 = g_strndup (start + 1, end - start - 1);

    if (end[1] == '<')
    {
	start = end + 1;
	if (strncmp (start + 2, "voice", 5))
	{
	    end = strstr (start + 6, ">");
	    item->voice = g_strndup (start + 7, end - start - 7);
	}
    }

    return item;
}



static SRCPresentationItem*
src_presentation_get_presentation_item_from_cond (const gchar *start,
						  const gchar *end,
						  const gchar *device)
{
    SRCPresentationItem *item = NULL;
    gchar *cond;
    
    sru_assert (start && end && device && end > start);
    
    cond = g_strndup (start, end - start + 1);
    if (cond)
	item = src_presentation_get_presentation_item_from_cond_ (cond, device);
    g_free (cond);
    
    return item;
}


static SRCPresentationItem*
src_presentation_get_presentation_item_from_cond1 (const gchar *start,
						   const gchar *end,
						   const gchar *device)
{
    SRCPresentationItem *item = NULL; 

    sru_assert (start && end && device && end > start);

    item = src_presentation_get_presentation_item_from_cond (start, end, device);
    if (item)
	item->type = SRC_PRESENTATION_INFO_TYPE_COND1;

    return item;
}

static SRCPresentationItem*
src_presentation_get_presentation_item_from_info_ (const gchar *info,
						   const gchar *device)
{
    gchar *start, *end;
    SRCPresentationItem *item;
    gchar *tmp;

    sru_assert (info && device);
    
    item = src_presentation_item_new ();
    if (!item)
        return NULL;

    item->type  = SRC_PRESENTATION_INFO_TYPE_INFO;
    
    end = strstr (info, "<");
    if (!end)
	end = (gchar*)info + strlen (info);    
    tmp = g_strndup (info, end - info);
    if (!tmp)
	return NULL;
    item->info = src_presentation_get_item_type (tmp);
    g_free (tmp);
    
    if (end[0] == '<')
    {
	start = end + 1;
	if (strncmp (start + 1, "voice", 5))
	{
	    end = strstr (start + 6, ">");
	    item->voice = g_strndup (start + 6, end - start - 6);
	}
    }

    return item;
}

static SRCPresentationItem*
src_presentation_get_presentation_item_from_info (const gchar *start,
						  const gchar *end,
						  const gchar *device)
{
    SRCPresentationItem *item = NULL;
    gchar *info;
    
    sru_assert (start && end && device && end > start);
    
    info = g_strndup (start, end - start + 1);
    if (info)
	item = src_presentation_get_presentation_item_from_info_ (info, device);
    g_free (info);
    
    return item;
}


static SRCPresentationItem*
src_presentation_get_presentation_item_from_paranthesis (const gchar *start,
							 const gchar *end,
							 const gchar *device)
{
    SRCPresentationItem *item;

    sru_assert (start && end && device && end > start);

    item = src_presentation_get_presentation_item_from_info (start, end, device);
    if (item)
	item->type = SRC_PRESENTATION_INFO_TYPE_PARANTHESIS;

    return item;
}


static gboolean 
src_presentation_get_presentation_items_from_formated_string
	    (const gchar *string, const gchar *device, SRCPresentationItem  ***presentation);


static SRCPresentationItem*
src_presentation_get_presentation_item_from_children_ (const gchar *children,
						       const gchar *device)
{
    gchar *start, *end;
    SRCPresentationItem *item;
    SRCPresentationItem **item1;
    gchar *tmp;

    sru_assert (children && device);

    item = src_presentation_item_new ();
    if (!item)
        return NULL;

    item->type  = SRC_PRESENTATION_INFO_TYPE_CHILDREN;

    /* 8 = strlen ("children") */
    start = (gchar*)children + 8;
    end = strstr (start, "]");
    tmp = g_strndup (start + 1, end - start - 1);
    if (!src_presentation_get_presentation_items_from_formated_string (tmp, device, &item1))
    {
	g_free (tmp);
	return FALSE;
    }
    g_free (tmp);

    item->data1 = item1;
        
    start = end + 1;
    end =   strstr (start, "]");
    if (end - start > 1)
	item->data2 = g_strndup (start + 3, end - start - 5);

    return item;
}


static SRCPresentationItem*
src_presentation_get_presentation_item_from_children (const gchar *start,
						      const gchar *end,
						      const gchar *device)
{
    SRCPresentationItem *item = NULL;
    gchar *children;
    
    sru_assert (start && end && device && end > start);
    
    children = g_strndup (start, end - start + 1);
    if (children)
	item = src_presentation_get_presentation_item_from_children_ (children, device);
    g_free (children);
    
    return item;
}


static gboolean
src_presentation_get_presentation_items_from_formated_string (const gchar *string,
							      const gchar *device,
							      SRCPresentationItem  ***presentation)
{
    gchar *crt;
    gchar *next;
    GSList *list = NULL;

    sru_assert (string && presentation);
    sru_assert (src_presentation_is_formated_string_ok (string, device));

    *presentation = NULL;
    crt = (gchar*) string;
    while (crt && crt[0])
    {
	SRCPresentationItem *item;

	if (crt[0] != '%')
	{
	    gchar *tmp;
	    tmp = strstr (crt, "%");
	    if (!tmp)
		tmp = crt + strlen (crt) - 1;
	    item = src_presentation_get_presentation_item_from_text (crt, tmp - 1, device);
	    crt = tmp;
	}
	else
	{
	    next = strstr (crt + 1, "%");

	    if (crt[1] == '!')
	    {
		item = src_presentation_get_presentation_item_from_special (crt + 2, next - 2, device);
	    }
	    else if (crt[1] == '?')
	    {
		item = src_presentation_get_presentation_item_from_cond (crt + 2, next - 2, device);
	    }
	    else if (crt[1] == '1')
	    {
		item = src_presentation_get_presentation_item_from_cond1 (crt + 2, next - 2, device);
	    }
	    else if (crt[1] == '(')
	    {
		item = src_presentation_get_presentation_item_from_paranthesis (crt + 2, next - 2, device);
	    }
	    else if (strncmp (crt + 1, "children", 8) == 0)
	    {
		next = strstr (crt, "]");
		next = strstr (next + 1, "]");
		next++;
		item = src_presentation_get_presentation_item_from_children (crt + 1, next - 1, device);
	    }
	    else
	    {
		item = src_presentation_get_presentation_item_from_info (crt + 1, next - 1, device);
	    }
	    crt = next + 1;
	}

	if (!item)
	{
	    return FALSE;
	}
	list = g_slist_append (list, item);
    }

    if (list)
    {	
	GSList *crt_list;
	gint i;
	*presentation = g_new (SRCPresentationItem *, g_slist_length (list) + 1);
	if (!(*presentation))
	    return FALSE;

	for (crt_list = list, i = 0; crt_list; crt_list = crt_list->next, i++)
	{
	    (*presentation)[i] = (SRCPresentationItem *) crt_list->data;
	}
	(*presentation)[i] = NULL;
	g_slist_free (list);
    }

    return *presentation ? TRUE : FALSE;    
}


static GHashTable *src_presentation_chunks_braille = NULL;
static GHashTable *src_presentation_chunks_speech  = NULL;


static gboolean
src_presentation_find_chunk_for_braille (const gchar *role,
					 const gchar *event,
					 SRCPresentationItem ***presentation)
{
    gchar *key;
    gchar *device = "braille";
    
    sru_assert (role && event && presentation);
    sru_assert (src_presentation_chunks_braille);

    /* fprintf (stderr, "\nFIND: %s %s %s", device, role, event); */
    
    *presentation = NULL;
    key = g_strconcat (role, "_", event, NULL);
    if (!key)
	return FALSE;

    *presentation = (SRCPresentationItem**) 
		    g_hash_table_lookup (src_presentation_chunks_braille, key);

    if (!(*presentation))
    {
	gchar *string;
	gboolean write_to_gconf = FALSE;
	string = src_presentation_try_read_formated_string_from_gconf (device, role, event);
	if (!string)
	{
	    string = src_presentation_try_read_formated_string_from_default (device, role, event);
	    write_to_gconf = TRUE;
	}
	if (string)
	{
	    if (src_presentation_is_formated_string_ok (string, device))
	    {
		SRCPresentationItem **presentation_;
		/* fprintf (stderr, "\nOK:%s:\n", string); */
		if (src_presentation_get_presentation_items_from_formated_string (string,
								    device,
								    &presentation_))
		{
		    g_hash_table_insert (src_presentation_chunks_braille, g_strdup (key), presentation_);
		}
		if (write_to_gconf)
		    src_presentation_write_formated_string_to_gconf (device, role, event, string);
	    
		*presentation = (SRCPresentationItem**) 
			g_hash_table_lookup (src_presentation_chunks_braille, key);
	    }
	    else
	    {
		fprintf (stderr, "\nNOT OK:%s:\n", string);
	    }
	    g_free (string);
	}
    }
    g_free (key);
    return *presentation ? TRUE : FALSE;
}


static gboolean
src_presentation_find_chunk_for_speech (const gchar *role,
					const gchar *event,
					SRCPresentationItem ***presentation)
{
    gchar *key;
    gchar *device = "speech";
    
    sru_assert (role && event && presentation);
    sru_assert (src_presentation_chunks_speech);

    /* fprintf (stderr, "\nFIND: %s %s %s", device, role, event); */
    
    *presentation = NULL;
    key = g_strconcat (role, "_", event, NULL);
    if (!key)
	return FALSE;

    *presentation = (SRCPresentationItem**) 
		    g_hash_table_lookup (src_presentation_chunks_speech, key);

    if (!(*presentation))
    {
	gchar *string;
	gboolean write_to_gconf = FALSE;
	string = src_presentation_try_read_formated_string_from_gconf (device, role, event);
	if (!string)
	{
	    string = src_presentation_try_read_formated_string_from_default (device, role, event);
	    write_to_gconf = TRUE;
	}
	if (string)
	{
	    if (src_presentation_is_formated_string_ok (string, device))
	    {
		SRCPresentationItem **presentation_;
		/* fprintf (stderr, "\nOK:%s:\n", string); */
		if (src_presentation_get_presentation_items_from_formated_string (string,
								    device,
								    &presentation_))
		{
		    g_hash_table_insert (src_presentation_chunks_speech, g_strdup (key), presentation_);
		}
		if (write_to_gconf)
		    src_presentation_write_formated_string_to_gconf (device, role, event, string);
	    }
	    else
		fprintf (stderr, "\nNOT OK:%s:\n", string);
	    *presentation = (SRCPresentationItem**) 
			g_hash_table_lookup (src_presentation_chunks_speech, key);
	    g_free (string);
	}
    }
    g_free (key);
    return *presentation ? TRUE : FALSE;
}

#if 0 /* this function is not used, but could be useful */
static gboolean
src_presentation_find_chunk_for_device (const gchar *device,
					const gchar *role,
					const gchar *event,
					SRCPresentationItem ***presentation)
{
    sru_assert (device && role && event && presentation);

    *presentation = NULL;
    
    if (strcmp (device, "braille") == 0)
	src_presentation_find_chunk_for_braille (role, event, presentation);
    else if (strcmp (device, "braille") == 0)
	src_presentation_find_chunk_for_speech (role, event, presentation);
    else
	sru_assert_not_reached ();

    return *presentation ? TRUE : FALSE;
}
#endif

gboolean
src_presentation_init ()
{
    src_presentation_chunks_speech = 
		g_hash_table_new_full (g_str_hash, g_str_equal,
				g_free, (GDestroyNotify)src_presentation_items_terminate);
    src_presentation_chunks_braille = 
		g_hash_table_new_full (g_str_hash, g_str_equal,
				g_free, (GDestroyNotify)src_presentation_items_terminate);

    if (!src_presentation_chunks_speech || !src_presentation_chunks_braille)
	return FALSE;

    return TRUE;
}


gboolean
src_presentation_terminate ()
{
    g_hash_table_destroy (src_presentation_chunks_braille);
    src_presentation_chunks_braille = NULL;
    g_hash_table_destroy (src_presentation_chunks_speech);
    src_presentation_chunks_speech = NULL;

    return TRUE;
}


static struct
{
    gchar	*role;
    gchar 	*braille_short;
    gchar	*braille_long;
    gchar 	*speech_short;
}src_presentation_role_pres[] =
    {
	{"unknown",			"UNK",	"UNKNOWN",		"Unknown",		},
	{"invalid",			"???",	"INVALID",		"Invalid", 		},
        {"alert", 			"ALR",	"ALERT",		"Alert",		},
	{"canvas",			"CNV",	"CANVAS",		"Canvas",		},
        {"check-box",			"CHK",	"CHECK_BOX",		"Check box",		},
	{"check-menu-item",		"MIT",	"MENU ITEM",		"Check Menu item",	},
	{"color-chooser",		"CCH",	"COLOR CHOOSER",	"Color chooser",	},
        {"column-header",		"CHD",	"COLUMN HEADER",	"Column header",	},
	{"combo-box",			"CBO",	"COMBO BOX",		"Combo box",		},
        {"desktop-icon",		"DIC",	"DESKTOP ICON",		"Desktop icon",		},
	{"desktop-frame",		"DFR",	"DESKTOP FRAME",	"Desktop frame",	},
        {"dialog",			"DLG",	"DIALOG",		"Dialog",		},
	{"directory-pane",		"DIP",	"DIRECTORY PANE",	"Directory pane",	},
        {"file-chooser",		"FCH",	"FILE CHOOSER",		"File chooser",		},
	{"filler",			"FLR",	"FILLER",		"Filler",		},
        {"frame",			"FRM",	"FRAME",		"Frame",		},
        {"glass-pane",			"GPN",	"GLASS PANE",		"Glass pane",		},
	{"HTML-container",		"HTM",	"HTML CONTAINER",	"H T M L container",	},
        {"icon",			"ICO",	"ICON",			"Icon",			},
        {"internal-frame",		"IFR",	"INTERNAL FRAME",	"Internal frame",	},
	{"label",			"LBL",	"LABEL",		"Label",		},
        {"layered-pane",		"LPN",	"LAYERED PANE",		"Layered pane",		},
	{"link",			"LNK",	"LINK",			"link",			},
	{"list",			"LST",	"LIST",			"List",			},
        {"list-item",			"LIT",	"LIST ITEM",		"List item",		},
	{"menu",			"MNU",	"MENU",			"Menu",			},
        {"menu-bar",			"MBR",	"MENU BAR",		"Menu bar",		},
	{"menu-item",			"MIT",	"MENU ITEM",		"Menu item",		},
        {"option-pane",			"OPN",	"OPTION PANE",		"Option pane",		},
        {"page-tab",			"PGT",	"PAGE TAB",		"Page tab",		},
	{"page-tab-list",		"PTL",	"PAGE TAB LIST",	"Page tab list",	},
        {"panel",			"PNL",	"PANEL",		"Panel",		},
        {"password-text",		"PWD",	"PASSWORD TEXT",	"Password text",	},
	{"popup-menu",			"PMN",	"POPUP MENU",		"Popup menu",		},
        {"progress-bar",		"PRG",	"PROGRESS BAR",		"Progress bar",		},
	{"push-button",			"PBT",	"PUSH BUTTON",		"Push button",		},
        {"radio-button",		"RAD",	"RADIO BUTTON",		"Radio button",		},
	{"radio-menu-item",		"MIT",	"MENU ITEM",		"Radio Menu item",	},
        {"root-pane",			"RPN",	"ROOT PANE",		"Root pane",		},
	{"row-header",			"RHD",	"ROW HEADER",		"Row header",		},
        {"scroll-bar",			"SCR",	"SCROLL BAR",		"Scroll bar",		},
	{"scroll-pane",			"SPN",	"SCROLL PANE",		"Scroll pane",		},
        {"separator",			"SEP",	"SEPARATOR",		"Separator",		},
	{"slider",			"SLD",	"SLIDER",		"Slider",		},
        {"split-pane",			"SPP",	"SPLIT PANE",		"Split pane",		},
	{"status-bar",			"STA",	"STATUS BAR",		"Status bar",		},
        {"table",			"TAB",	"TABLE",		"Table",		},
	{"table-cell",			"CEL",	"TABLE CELL",		"Table cell",		},
        {"table-column-header",		"TCH",	"TABLE COLUMN HEADER",	"Table column header",	},
	{"table-row-header",		"TRH",	"TABLE ROW HEADER",	"Table row header",	},
        {"text",			"TXT",	"TEXT",			"Text",			},
	{"toggle-button",		"TOG",	"TOGGLE BUTTON",	"Toggle button",	},
        {"tool-bar",			"TOL",	"TOOL BAR",		"Tool bar",		},
	{"tool-tip",			"TIP",	"TOOL TIP",		"Tool tip",		},
        {"tree",			"TRE",	"TREE",			"Tree",			},
        {"tree-item",			"TRI",	"TREE ITEM",		"Tree item",		},
	{"tree-table",			"TRT",	"TREE TABLE",		"Tree table",		},
	{"viewport",			"VWP",	"VIEWPORT",		"Viewport",		},
	{"window",			"WND",	"WINDOW",		"Window",		},
	{"accelerator-label",		"ACC",	"ACCELERATOR LABEL",	"Accelerator label",	},
        {"animation",			"ANI",	"ANIMATION",		"Animation",		},
	{"arrow",			"ARR",	"ARROW",		"Arrow",		},
        {"calendar",			"CAL",	"CALENDAR",		"Calendar",		},
	{"date-editor",			"DAT",	"DATE EDITOR",		"Date editor",		},
        {"dial",			"DIL",	"DIAL",			"Dial",			},
	{"drawing-area",		"DRW",	"DRAWING AREA",		"Drawing area",		},
        {"font-chooser",		"FNT",	"FONT CHOOSER",		"Font chooser",		},
	{"image",			"IMG",	"IMAGE",		"Image",		},
        {"spin-button",			"SPN",	"SPIN BUTTON",		"Spin button",		},
	{"terminal",			"TRM",	"TERMINAL",		"Terminal",		},
        {"extended",			"EXT",	"EXTENDED",		"Extended",		},
	{"table-line",			"TLI",	"TABLE LINE",		"Table line",		},
        {"table-columns-header",	"TCH","TABLE COLUMNS HEADER",	"Table columns header",	},
	{"title-bar",			"TIT",	"TITLE BAR",		"Title bar",		}
    };

static gint
src_presentation_get_role_index (const gchar *role)
{
    gint i;

    sru_assert (role);

    for (i = 0; i < G_N_ELEMENTS (src_presentation_role_pres); i++)
	if (strcmp (src_presentation_role_pres[i].role, role) == 0)
	    return i;
    
    return 0;
}




static G_CONST_RETURN gchar* 
src_presentation_get_role_name_for_braille (const gchar *role)
{
    gint index;
    sru_assert (role);

    index = src_presentation_get_role_index (role);

    return src_presentation_role_pres[index].braille_short;
}

static G_CONST_RETURN gchar* 
src_presentation_get_role_name_for_speech (const gchar *role)
{
    gint index;

    sru_assert (role);

    index = src_presentation_get_role_index (role);

    return src_presentation_role_pres[index].speech_short;
}


G_CONST_RETURN gchar* 
src_presentation_get_role_name_for_device (const gchar *role,
					   const gchar *device)
{
    if (strcmp (device, "braille") == 0)
	return src_presentation_get_role_name_for_braille (role);
    else if (strcmp (device, "speech") == 0)
	return src_presentation_get_role_name_for_speech (role);
    else
	sru_assert_not_reached ();

    return NULL;
}



static gboolean
src_presentation_get_info_from_sro (SRObject *obj,
				    SRCPresentationItemType info_type,
				    const gchar *device,
				    gpointer info)
{
    gboolean rv = FALSE;

    sru_assert (obj && info && device);

    if (info)
	*((gchar**)info) = NULL;

    switch (info_type)
    {
	case SRC_PRESENTATION_LEVEL:
	    {
		SRLong tmp;
		rv = sro_tree_item_get_level (obj, &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{	
		    *((gchar **)info) = g_strdup_printf ("%ld", tmp);
		}
	    }
	    break;
	case SRC_PRESENTATION_NAME:
	    {
		gchar *tmp;
		rv = sro_get_name (obj, &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{	
		    *((gchar **)info) = g_strdup (tmp);
		    SR_freeString (tmp);
		}
	    }
	    break;
	case SRC_PRESENTATION_ROLE:
	    {
		gchar *tmp;
		sro_get_role_name (obj, &tmp, SR_INDEX_CONTAINER);
		
		if (tmp)
		{
		    G_CONST_RETURN gchar *tmp2;
		    
		    tmp2 = src_presentation_get_role_name_for_device (tmp, device);
		    if (tmp2)
		    {
			*((gchar **)info) = g_strdup (tmp2);
			rv = TRUE;
		    }
		    SR_freeString (tmp);
		}
	    }    
	    break;
	case SRC_PRESENTATION_ACCELERATOR:
	    {
		gchar *tmp;
		rv = sro_get_accelerator (obj, &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{
		    *((gchar **)info) = g_strdup (tmp);
		    SR_freeString (tmp);
		}
	    }    
	    break;
	case SRC_PRESENTATION_SHORTCUT:
	    {
		gchar *tmp;
		rv = sro_get_shortcut (obj, &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{
		    *((gchar **)info) = g_strdup (tmp);
		    SR_freeString (tmp);
		}
	    }    
	    break;
	case SRC_PRESENTATION_STATE_CHECKED:
	    {
		SRState state;
		if (sro_get_state (obj, &state, SR_INDEX_CONTAINER))
		    rv = state & SR_STATE_CHECKED ? TRUE : FALSE;
	    }
	    break;
	case SRC_PRESENTATION_STATE_UNCHECKED:
	    {
		SRState state;
		if (sro_get_state (obj, &state, SR_INDEX_CONTAINER))
		    rv = (state & SR_STATE_CHECKABLE) && !(state & SR_STATE_CHECKED) ? TRUE : FALSE;
	    }
	    break;
	case SRC_PRESENTATION_STATE_EXPANDED:
	    {
		SRState state;
		if (sro_get_state (obj, &state, SR_INDEX_CONTAINER))
		    rv = state & SR_STATE_EXPANDED ? TRUE : FALSE;
	    }
	    break;
	case SRC_PRESENTATION_STATE_COLLAPSED:
	    {
		SRState state;
		if (sro_get_state (obj, &state, SR_INDEX_CONTAINER))
		    rv = (state & SR_STATE_EXPANDABLE) && !(state & SR_STATE_EXPANDED) ? TRUE : FALSE;
	    }
	    break;
	case SRC_PRESENTATION_STATE_MINIMIZED:
	    {
		SRState state;
		if (sro_get_state (obj, &state, SR_INDEX_CONTAINER))
		    rv = (state & SR_STATE_MINIMIZED) ? TRUE : FALSE;
	    }
	    break;

	case SRC_PRESENTATION_STATE_ENABLED:
	    {
		SRState state;
		if (sro_get_state (obj, &state, SR_INDEX_CONTAINER))
		    rv = state & SR_STATE_ENABLED;
	    }
	    break;
	

	case SRC_PRESENTATION_CHILDREN_COUNT:
	    {
		guint32 cc;
		if (!sro_get_children_count (obj, &cc))
		    cc = 0;
		*((gchar **)info) = g_strdup_printf ("%d", cc);
		rv = TRUE;
	    }
	    break;
	case SRC_PRESENTATION_TEXT_LINE:
	    {
		gchar *tmp;
		rv = sro_text_get_text_from_caret (obj, SR_TEXT_BOUNDARY_LINE,
					    &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{
		    *((gchar **)info) = g_strdup (tmp);
		    SR_freeString (tmp);
		}
	    }
	    break;
	case SRC_PRESENTATION_TEXT_BEST:
	    {
		static SRLong last_offset = 0, last_offset2 = 0;
		SRLong crt_offset;
		SRTextBoundaryType type;
		gchar *tmp;
		sro_text_get_abs_offset (obj, &crt_offset, SR_INDEX_CONTAINER);
		
		type = SR_TEXT_BOUNDARY_LINE;
		
		if (last_offset == crt_offset)
		    last_offset = last_offset2;
		if (abs (crt_offset - last_offset) == 1)
		    type = SR_TEXT_BOUNDARY_CHAR;
		else if (sro_text_is_same_line (obj, last_offset, SR_INDEX_CONTAINER))
		    type = SR_TEXT_BOUNDARY_WORD;
		else
		    type = SR_TEXT_BOUNDARY_LINE;
		if (sro_text_get_text_from_caret (obj, type,
				&tmp, SR_INDEX_CONTAINER))
		    {
			*((gchar **)info) = g_strdup (tmp);
			rv = TRUE;
			SR_freeString (tmp);
		    }
		last_offset2 = last_offset;
		last_offset = crt_offset;
	    }
	    break;
	case SRC_PRESENTATION_TEXT_CHAR:
	    {
		gchar *tmp;
		rv = sro_text_get_text_from_caret (obj, SR_TEXT_BOUNDARY_CHAR,
					    &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{
		    *((gchar **)info) = g_strdup (tmp);
		    SR_freeString (tmp);
		}
	    }
	    break;
	case SRC_PRESENTATION_TEXT_SELECTION:
	    {
		gchar **tmp;
		rv = sro_text_get_selections (obj, &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{
		    *((gchar **)info) = g_strdup (tmp[0]);
		    SR_strfreev (tmp);
		}
	    }
	    break;
	case SRC_PRESENTATION_TEXT_DIFFERENCE:
	    {
		gchar *tmp;
		rv = sro_text_get_difference (obj, &tmp, SR_INDEX_CONTAINER);
		if (rv)
		{
		    *((gchar **)info) = g_strdup (tmp);
		    SR_freeString (tmp);
		}
	    }
	    break;
	case SRC_PRESENTATION_VALUE_CRT:
	    {
		gdouble val;
		rv = FALSE;
		if (sro_value_get_crt_val (obj, &val, SR_INDEX_CONTAINER))
		{
		    *((gchar **)info) = g_strdup_printf ("%.0f", val);
		    rv = TRUE;
		}
	    }
	    break;
	default:
	    fprintf (stderr, "\n%d", info_type);/* removable */
	    sru_assert_not_reached ();	
	    break;
    }
    return rv;
}


static gchar*
src_presentation_present_sro_for_device_ (SRObject *obj,
					 const gchar *device,
					 SRCPresentationItem **chunk)
{
    gint i;
    gchar *rv;
    gboolean has_voice;
    sru_assert (obj && chunk && *chunk);

    has_voice = strcmp (device, "speech") == 0 ? TRUE : FALSE;
    i = 0;
    rv = g_strdup ("");
    while (chunk[i])
    {
	gchar *add = NULL;

	switch (chunk[i]->type)
	{
	    case SRC_PRESENTATION_INFO_TYPE_INFO:
		{
		    gchar *info;
		    if (src_presentation_get_info_from_sro (obj, chunk[i]->info, device, &info))
		    {
			if (info)
			{
			    add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, info);
			    g_free (info);
			}
		    }
		}
		break;
	    case SRC_PRESENTATION_INFO_TYPE_CHILDREN:
		{
		    guint32 cnt, j;
		    
		    add = g_strdup ("");		    
		    sro_get_children_count (obj, &cnt);
		    for (j = 0; j < cnt; j++)
		    {
			SRObject *child;
			gchar *tmp, *tmp2, *tmp3;

			if (!sro_get_i_child (obj, j, &child))
			    continue;
			
			tmp = add;
			tmp2 = src_presentation_present_sro_for_device_ (child, device, (SRCPresentationItem **)chunk[i]->data1);
			
			tmp3 = NULL;
			if (j != cnt - 1 && chunk[i]->data2)
			{
			    gchar *tmp4;
			    tmp4 = g_strdup_printf ("dot%s", (gchar *)chunk[i]->data2);
			    if (tmp4)
			    {
				tmp3 = src_xml_make_part ("DOTS", NULL, tmp4);
				g_free (tmp4);
			    }
			}
			add = g_strconcat (add, tmp2, tmp3, NULL);
			g_free (tmp);
			g_free (tmp2);
			g_free (tmp3);
			sro_release_reference (child);
		    }
		}
		break;
	    case SRC_PRESENTATION_INFO_TYPE_COND:
		{
		    gchar *info;
		    if (src_presentation_get_info_from_sro (obj, chunk[i]->info, device, &info))
		    {
			add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, (gchar *)chunk[i]->data1);
			if (info)
			    g_free (info);
		    }
		    else
		    {
			add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, (gchar *)chunk[i]->data2);
		    }
		}
		break;
	    case SRC_PRESENTATION_INFO_TYPE_PARANTHESIS:
		{
		    gchar *info, *info2;
		    info2 = NULL;
		    if (src_presentation_get_info_from_sro (obj, chunk[i]->info, device, &info))
		    {
			info2 = g_strconcat ("(", info, ")", NULL);
			g_free (info);
		    }
		    
		    if (info2)
		    {
			add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, info2);
			g_free (info2);
		    }
		}
		break;
	    case SRC_PRESENTATION_INFO_TYPE_COND1:
		{
		    gchar *info;
		    if (src_presentation_get_info_from_sro (obj, chunk[i]->info, device, &info))
		    {
			if (info && strcmp (info, "1") ==0)
			    add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, (gchar *)chunk[i]->data1);
			else
			    add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, (gchar *)chunk[i]->data2);
			if (info)
			    g_free (info);
		    }
		    else
		    {
			add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, (gchar *)chunk[i]->data2);
		    }
		}
		break;
	    case SRC_PRESENTATION_INFO_TYPE_TEXT:
		if (chunk[i]->data1)
		    add = src_xml_format ("TEXT", has_voice ? src_speech_get_voice ((gchar *)chunk[i]->voice) : NULL, (gchar *)chunk[i]->data1);
		break; 
	    case SRC_PRESENTATION_INFO_TYPE_SPECIAL:
		{
		    if (chunk[i]->data1)
		    {
			gchar *tmp;
			tmp = g_strdup_printf ("dot%s", (gchar *)chunk[i]->data1);
			if (tmp)
			{
			    add = src_xml_make_part ("DOTS", NULL, tmp);
			    g_free (tmp);
			}
		    }
		}
		break;
	    default:
		sru_assert_not_reached ();
	}
	if (add)
	{
	    gchar *tmp = rv;
	    if (strcmp (device, "braille") == 0)
	    {
		rv = g_strconcat (rv, rv[0] != '\0' ? "<TEXT> </TEXT>" : "", add, NULL); 
	    }
	    else
	    {
		rv = g_strconcat (rv, add, NULL);
	    }
	    g_free (tmp);
	}
	i++;
    }
    /*fprintf (stderr, "\n%s", rv);*/
    return rv;
}

#if 0
gchar*
src_presentation_present_sro_for_magnifier (SRObject *obj, 
					    const gchar *reason)
{
    gchar *presentation_ = NULL;
    SRRectangle location;
    gboolean ok = FALSE;
    SRLong index;
    SRObjectRoles role;

    sru_assert (obj && reason);
    
    sro_get_role (obj, &role, SR_INDEX_CONTAINER);
    if ((role == SR_ROLE_TEXT ) &&
	sro_text_get_caret_offset (obj, &index, SR_INDEX_CONTAINER) &&
	sro_text_get_location_at_index (obj, index, &location, SR_INDEX_CONTAINER))
    {
	ok = TRUE;
    }
    else if (sro_get_location (obj, SR_COORD_TYPE_SCREEN, &location, SR_INDEX_CONTAINER))
    {
	ok = TRUE;
    }
    if (ok)
    {
	presentation_ = g_strdup_printf ("ROILeft =\"%d\" ROITop =\"%d\" ROIWidth =\"%d\" ROIHeight=\"%d\"",
					location.x,
					location.y,
					location.x + location.width,
					location.y + location.height);
    }
    
    return presentation_;
}
#endif

gchar*
src_presentation_present_sro_for_magnifier (SRObject *obj, 
					    const gchar *reason)
{
    gchar *role;
    gchar *device = "magnifier";
    gchar *presentation_ = NULL;

    sru_assert (obj && reason);

    if (sro_get_role_name (obj, &role, SR_INDEX_CONTAINER))
    {
	gchar *string;
	gboolean write_to_gconf = FALSE;
	string = src_presentation_try_read_formated_string_from_gconf (device, role, reason);
	if (!string)
	{
	    string = src_presentation_try_read_formated_string_from_default (device, role, reason);
	    write_to_gconf = TRUE;
	}
	if (string)
	{
	    if (src_presentation_is_formated_string_ok (string, device))
	    {
		/* fprintf (stderr, "\nOK:%s:\n", string); */
		/* if (write_to_gconf)
		    src_presentation_write_formated_string_to_gconf (device, role, event, string);
		*/
	    }
	    else
	    {
		/* [Cfprintf (stderr, "\nNOT OK:%s:\n", string);*/
	    }
	    
	    {
		SRRectangle location;
		gboolean ok = TRUE;
		if (strncmp (string + 1, "text:line", 9) == 0)
		{
		     sro_text_get_text_location_from_caret (obj, SR_TEXT_BOUNDARY_LINE,
	    	    	    SR_COORD_TYPE_SCREEN, &location, SR_INDEX_CONTAINER);
		}
		else if (strncmp (string + 1, "text:char", 9) == 0)
		{
		    SRLong index;
		    
		    sro_text_get_caret_offset (obj, &index, SR_INDEX_CONTAINER);
		    sro_text_get_location_at_index (obj, index, &location, SR_INDEX_CONTAINER);
		}
		else if (strncmp (string + 1, "object", 6) == 0)
		{
		    sro_get_location (obj, SR_COORD_TYPE_SCREEN, &location, SR_INDEX_CONTAINER);
		}
		else
		    ok = FALSE;
		if (ok)
		{
		    gchar *tmp1, *tmp2, *alg;
		    tmp1 = strstr (string, "<");
		    tmp2 = strstr (string, ">");
		    sru_assert (tmp1 && tmp2);
		    alg = g_strndup (tmp1 + 1, tmp2 - tmp1 - 1);
		    presentation_ = g_strdup_printf ("ROILeft =\"%d\" ROITop =\"%d\" ROIWidth =\"%d\" ROIHeight=\"%d\" %s",
					location.x,
					location.y,
					location.x + location.width,
					location.y + location.height,
					alg);
		    g_free (alg);
		}
	    }
	    g_free (string);
	}
    	SR_freeString (role);    
    }
    return presentation_;
}

gchar*
src_presentation_present_sro_for_speech (SRObject *obj, 
					 const gchar *reason)
{
    SRCPresentationItem **presentation;
    gchar *role;
    gchar *device = "speech";
    gchar *presentation_ = NULL;

    sru_assert (obj && reason);

    if (sro_get_role_name (obj, &role, SR_INDEX_CONTAINER))
    {
	if (src_presentation_find_chunk_for_speech (role, reason, &presentation))
	{
	    presentation_ = src_presentation_present_sro_for_device_ (obj, device, presentation);
	}
	SR_freeString (role);    
    }

    return presentation_;
}

gchar*
src_presentation_present_sro_for_braille (SRObject *obj, 
					  const gchar *reason)
{
    SRCPresentationItem  **presentation;
    gchar *role;
    gchar *device = "braille";
    gchar *presentation_ = NULL;

    sru_assert (obj && reason);

    if (sro_get_role_name (obj, &role, SR_INDEX_CONTAINER))
    {
	if (src_presentation_find_chunk_for_braille (role, reason, &presentation))
	{
	    presentation_ = src_presentation_present_sro_for_device_ (obj, device, presentation);
	}
	SR_freeString (role);    
    }

    return presentation_;
}


gchar *
src_presentation_present_sro_for_device (SRObject *obj, 
					 const gchar *reason, 
					 const gchar *device)
{
    gchar *presentation_ = NULL;
    
    sru_assert (obj && device && reason);

    if (strcmp (device, "magnifier") == 0)
	presentation_ = src_presentation_present_sro_for_magnifier (obj, reason);
    else  if (strcmp (device, "speech") == 0)
	presentation_ = src_presentation_present_sro_for_speech (obj, reason);
    else if (strcmp (device, "braille") == 0)
	presentation_ = src_presentation_present_sro_for_braille (obj, reason);
    else
	sru_assert_not_reached ();

    return presentation_;
}


G_CONST_RETURN gchar* 
src_pres_get_role_name_for_speech (const gchar *role)
{
    gint index;

    sru_assert (role);

    index = src_presentation_get_role_index (role);

    return src_presentation_role_pres[index].speech_short;
}
