/* srspc.c
 *
 * Copyright 2001, 2002 Sun Microsystems, Inc.,
 * Copyright 2001, 2002 BAUM Retec, A.G.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <ctype.h>

#include "srmain.h"
#include "srspc.h"
#include "../speech/libsrs/srsxmlapi.h"
#include "SREvent.h"
#include "srpres.h"
#include "libsrconf.h"
#include "srintl.h"

#include "sys/types.h"
#include "sys/stat.h"
#include "fcntl.h"
#include "unistd.h"

#define STR(X) (X == NULL? "" : X)

/* FIXME:next 4 variables must be static */
SRCSpeechTextEchoType  	   src_speech_text_echo_type     = SRC_SPEECH_TEXT_ECHO_CHAR;
SRCSpeechModifiersEchoType src_speech_modifiers_echo_type= SRC_SPEECH_MODIFIERS_ECHO_ALL;
SRCSpeechSpacesEchoType    src_speech_spaces_echo_type   = SRC_SPEECH_SPACES_ECHO_ALL;
SRCSpeechCursorsEchoType   src_speech_cursors_echo_type  = SRC_SPEECH_CURSORS_ECHO_ALL;
SRCSpeechDictionaryType    src_speech_dictionary_type    = SRC_SPEECH_DICTIONARY_YES;

static SRCSpeechSpellingMode      src_speech_spelling_mode      = SRC_SPEECH_SPELL_NORMAL;
static SRCSpeechCountMode     	  src_speech_count_mode         = SRC_SPEECH_COUNT_NONE;
static gchar 	**src_speech_engine_voices = NULL;
static gchar 	**src_speech_voices        = NULL;
static gboolean src_speech_initialized     = FALSE;

static GHashTable *src_dictionary_hash 	= NULL;


static gchar**
src_speech_get_engine_voices ()
{
    gint index, index2;
    gint count = -1;
    gchar **rv;

    sru_assert (src_speech_initialized);

    srconf_get_data_with_default (SPEECH_VOICE_COUNT, CFGT_INT, 
				&count, &count, SPEECH_PARAMETER_SECTION);

    if (count <= 0)
	return NULL;
    
    rv = g_new (gchar*, count + 1);
    index2 = 0;
    for (index = 0; index < count; index++)
    {
	gchar *path, *tmp;
	path = g_strdup_printf ("voice_%d_name", index);
	srconf_get_data_with_default (path, CFGT_STRING, &tmp, NULL,
					    SPEECH_PARAMETER_SECTION);
	if (tmp && tmp[0])
	{
	    rv[index2] = g_strdup (tmp);
/*	     fprintf (stderr, "\nENGINE VOICE[%d]=%s", index2, rv[index2]);*/
	    index2++;
	}
	g_free (tmp);
	g_free (path);
    }
    rv[index2] = NULL;

    return rv;
}


static gboolean
src_speech_valid_engine_voice (gchar *voice)
{
    gint i;

    sru_assert (src_speech_initialized);
    sru_assert (src_speech_engine_voices);
    sru_assert (voice);

    for (i = 0; src_speech_engine_voices[i]; i++)
    {
	if (strcmp (voice, src_speech_engine_voices[i]) == 0)
	    return TRUE;
    }
    return FALSE;
}


static gchar**
src_speech_get_voices ()
{
    gchar **rv = NULL;
    gint cnt, index;
    GSList *voices = NULL;

    sru_assert (src_speech_initialized);

    srconf_get_data_with_default (SPEECH_GNOPERNICUS_SPEAKERS, CFGT_LIST,
				&voices, NULL, SPEECH_PARAMETER_SECTION);

    cnt = voices ? g_slist_length (voices) + 2 : 2;
    rv = g_new (gchar*, cnt);
    index = 0;

    if (voices)
    {
	GSList *crt = NULL;
	for (crt = voices; crt; crt = crt->next)
	{
	    gchar *voice = crt->data;
	    if (voice && voice[0])
	    {
		rv[index] = voice;
		crt->data = NULL;
/*		 fprintf (stderr, "\nGNOPERNICUS VOICE[%d]=%s", index, rv[index]);*/
	        index++;
	    }
	}
    }
    g_slist_free (voices);
    rv[index] = g_strdup ("default");
    rv[index + 1] = NULL;

    return rv;
}


static gboolean
src_speech_valid_voice (gchar *voice)
{
    gint i;

    sru_assert (src_speech_initialized);
    sru_assert (src_speech_voices);
    sru_assert (voice);

    for (i = 0; src_speech_voices[i]; i++)
    {
	if (strcmp (voice, src_speech_voices[i]) == 0)
	    return TRUE;
    }    
    return FALSE;
}


#define src_speech_set_string_from_list_element(list_el, string)	      \
	    {								      \
		if (list_el && list_el->data && ((gchar*)(list_el->data))[0]) \
		{							      \
		    *string = list_el->data;				      \
		    list_el->data = NULL;				      \
		}							      \
		else							      \
		{							      \
		    *string = NULL;					      \
		}							      \
	    }


static gboolean
src_speech_get_voice_parameters (gchar *voice,
				 gchar **name,
				 gchar **volume,
				 gchar **rate,
				 gchar **pitch)
{
    GSList *voices = NULL;

    sru_assert (src_speech_initialized);
    sru_assert (voice && voice[0]);

    srconf_get_data_with_default (voice, CFGT_LIST, &voices, 
					    NULL, SPEECH_PARAMETER_SECTION);

    if (voices)
    {
	GSList *crt = voices;

	if (name)
	    src_speech_set_string_from_list_element (crt, name);

	if (crt)
	    crt = crt->next;
	if (volume)
	    src_speech_set_string_from_list_element (crt, volume);

	if (crt)
	    crt = crt->next;
	if (rate)
	    src_speech_set_string_from_list_element (crt, rate);

	if (crt)
	    crt = crt->next;
	if (pitch)
	    src_speech_set_string_from_list_element (crt, pitch);

	srconf_free_slist (voices);
    }
    else
    {
	if (name)
	    *name = NULL;
	if (volume)
	    *volume = NULL;
	if (rate)
	    *rate = NULL;
	if (pitch)
	    *pitch = NULL;
    }
    /* fprintf (stderr, "\ngotten parameters for voice: %s are name: %s volume: %s rate: %s pitch: %s",
		    voice,
		    *name ? *name : "(null)",
		    *volume ? *volume : "(null)",
		    *rate ? *rate : "(null)",
		    *pitch ? *pitch : "(null)"); */
    return TRUE;
}


static gboolean
src_speech_set_voice_parameters (gchar *voice,
				 gchar *name_,
				 gchar *volume_,
				 gchar *rate_,
				 gchar *pitch_)
{
    gchar *name, *volume, *rate, *pitch;
    gboolean rv = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (voice && voice[0]);

    src_speech_get_voice_parameters (voice, &name, &volume, &rate, &pitch);

    if (name_)
    {
	g_free (name);
	name = g_strdup (name_);
    }
    if (volume_)
    {
	g_free (volume);
	volume = g_strdup (volume_);
    }
    if (rate_)
    {
	g_free (rate);
	rate = g_strdup (rate_);
    }
    if (pitch_)
    {
	g_free (pitch);
	pitch = g_strdup (pitch_);
    }
    if (! (name && name[0] && volume && volume[0] && rate && rate[0] 
		&& pitch && pitch[0]))
    {
	sru_warning (_("incorect parameters for voice \"%s\""), voice); 
    }
    else
    {
        GSList *list = NULL;
	list = g_slist_append (list, name);
	list = g_slist_append (list, volume);
	list = g_slist_append (list, rate);
	list = g_slist_append (list, pitch);
	srconf_set_data (voice, CFGT_LIST, list, SPEECH_PARAMETER_SECTION);
	srconf_free_slist (list);
	rv = TRUE;
    }    

    return rv;
}


static gboolean
src_speech_send (gchar *spcoutput)
{
    gboolean rv = FALSE;
    /*FIXME: busy must not be here */
    static gboolean busy = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (spcoutput && spcoutput[0]);
    
    if (busy)
	return FALSE;
    busy = TRUE;
    /*FIXME: next line must be valid, no folowing 2 */
    /* rv = srs_xml_output (spcoutput, strlen(spcoutput)); */
    srs_xml_output (spcoutput, g_utf8_strlen(spcoutput, -1));
    rv = TRUE;
    
/*     fprintf (stderr, "\n%s", spcoutput); */

    busy = FALSE;

    if (!rv)
	sru_warning (_("failed to send to speech next string \"%s\""),
					    spcoutput);

    return rv;
}


static gboolean
src_speech_create_voice (gchar *voice,
			 gchar *name_,
			 gchar *volume_,
			 gchar *rate_,
			 gchar *pitch_)
{
    gchar *voice_;
    gchar *name, *volume, *rate, *pitch;

    gboolean rv = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (voice);

    name = volume = rate = pitch = NULL;
    if (name_)
    {
	sru_assert (src_speech_valid_engine_voice (name_));
	name = g_strconcat (" TTSVoice=\"", name_, "\"", NULL);
    }
    if (rate_)
	rate = g_strconcat (" rate=\"", rate_, "\"", NULL);
    if (volume_)
	volume = g_strconcat (" volume=\"", volume_, "\"", NULL);
    if (pitch_)
	pitch = g_strconcat (" pitch=\"", pitch_, "\"", NULL);

    voice_ = g_strconcat (	"<SRSOUT>"
				    "<VOICE"
					" ID=\"", voice, "\""
					" TTSEngine=\"gnome-speech\"",
					/* " priority=\"0\"" */
					/* " preempt=\"yes\"", */
					name ? name : "",
					rate ? rate : "",
					pitch ? pitch : "",
					volume ? volume : "",
				    "></VOICE>"
				"</SRSOUT>",
				NULL);

    if (voice_)
	rv = src_speech_send (voice_);
    if (!rv)
	sru_warning (_("could not create/modify voice \"%s\""), voice);
    g_free (voice_);

    return rv;
}


gboolean
src_speech_init_voice (gchar *voice)
{
    gchar *name, *volume, *rate, *pitch;
    gboolean rv = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (voice && voice[0]);
    
    name = volume = rate = pitch = NULL;
    if (src_speech_get_voice_parameters (voice, &name, &volume, &rate, &pitch))
    {
	if (!name || !src_speech_valid_engine_voice (name))
	{
	    g_free (name);
	    name = g_strdup (src_speech_engine_voices[0]);
	    src_speech_set_voice_parameters (voice, name, volume, rate, pitch);
	}
	rv = src_speech_create_voice (voice, name, volume, rate, pitch);
    }

    g_free (name);
    g_free (volume);
    g_free (rate);
    g_free (pitch);

    return rv;
}

/*FIXME: remove _ from function name */
static gboolean 
src_speech_say_ (gchar *text_,
		gchar *voice,
		SRCSpeechSpellingMode smode,
		SRCSpeechPriority priority)
{
    gchar *text, *attr;
    gboolean rv = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (text_ && text_[0] && voice);

    text = attr = NULL;
    attr = src_speech_get_text_voice_attributes (voice, smode);
    if  (attr)
	text = src_xml_format ("TEXT", attr, text_);
    if (text)
	rv = src_speech_send_chunk (text, priority);
    g_free (text);
    g_free (attr);

    return rv;
}


gboolean 
src_speech_say_message (gchar *message)
{
    gboolean rv = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (message && message[0]);

    rv = src_speech_say_ (message, "message", SRC_SPEECH_SPELL_AUTO, 
				SRC_SPEECH_PRIORITY_MESSAGE);

    return rv;
}


static gboolean
src_speech_add_voice (gchar *voice)
{
    gchar **voices;
    gint i;

    sru_assert (src_speech_initialized);
    sru_assert (voice);
    sru_assert (src_speech_voices);

    for (i = 0; src_speech_voices[i]; i++)
	if (strcmp (voice, src_speech_voices[i]) == 0)
	    return TRUE;
    voices = src_speech_voices;
    src_speech_voices = g_new (gchar*, i + 2);
    for (i = 0; voices[i]; i++)
	src_speech_voices[i] = voices[i];
    src_speech_voices[i] = g_strdup (voice);
    src_speech_voices[i + 1] = NULL;
    g_free (voices);

    return TRUE;
}


static gboolean
src_speech_remove_voice (gchar *voice)
{
    gboolean rv = FALSE;
    gint i;

    sru_assert (src_speech_initialized);
    sru_assert (voice);
    sru_assert (src_speech_voices);

    for (i = 0; src_speech_voices[i]; i++)
	if (strcmp (voice, src_speech_voices[i]) == 0)
	    break;
    if (src_speech_voices[i])
	rv = TRUE;
    for ( ; src_speech_voices[i]; i++)
	src_speech_voices[i] = src_speech_voices[i + 1];

    return rv;
}


gboolean
src_speech_test_voice (gchar *voice)
{
    gchar *name, *volume, *rate, *pitch;
    gboolean rv = FALSE;

    sru_assert (src_speech_initialized);
    sru_assert (voice && voice[0]);
    sru_assert (src_speech_voices);
    
    if (!src_speech_add_voice ("test"))
	return FALSE;

    name = volume = rate = pitch = NULL;
    if (src_speech_get_voice_parameters (voice, &name, &volume, &rate, &pitch))
    {
	if (name && src_speech_valid_engine_voice (name))
	    rv = src_speech_create_voice ("test", name, volume, rate, pitch);
    }

    if (rv)
    {
	gchar *test;
	test = g_strdup_printf (_("voice %s will sound this way"), voice);
	rv = src_speech_say_ (test, "test", SRC_SPEECH_SPELL_NORMAL, TRUE);
	g_free (test);
    }
    if (!rv)
	sru_warning (_("could not test voice \"%s\""), voice);

    src_speech_remove_voice ("test");
    g_free (name);
    g_free (volume);
    g_free (rate);
    g_free (pitch);

    return rv;
}


static gboolean
src_speech_voices_init ()
{
    gchar ** voices;

    sru_assert (src_speech_initialized);

    voices = src_speech_voices;
    if (voices)
    {
	gint i;
	for (i = 0; voices[i]; i++)
	    src_speech_init_voice (voices[i]);
    }

    return TRUE;
}


static void
src_speech_voices_terminate ()
{
    sru_assert (src_speech_initialized);
}


gboolean
src_speech_set_spelling_mode (SRCSpeechSpellingMode mode)
{
    sru_assert (src_speech_initialized);
    sru_assert (mode  != SRC_SPEECH_SPELL_AUTO);

    src_speech_spelling_mode = mode;

    return TRUE;
}


SRCSpeechSpellingMode
src_speech_get_spelling_mode ()
{
    sru_assert (src_speech_initialized);

    return src_speech_spelling_mode;
}


gboolean
src_speech_set_count_mode (SRCSpeechCountMode mode)
{
    sru_assert (src_speech_initialized);
    sru_assert (mode  != SRC_SPEECH_COUNT_AUTO);

    src_speech_count_mode = mode;
    
    return TRUE;
}


static gboolean
src_speech_set_count_mode_from_string (gchar *mode)
{
    gboolean rv = TRUE;
    SRCSpeechCountMode mode_ = src_speech_count_mode;

    sru_assert (mode);

    if (strcmp (mode, "ALL") == 0)
	mode_ = SRC_SPEECH_COUNT_ALL;
    else if (strcmp (mode, "NONE") == 0)
	mode_ = SRC_SPEECH_COUNT_NONE;
    else
    {
	sru_warning (_("value not supported for speech count mode \"%s\""), mode);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_count_mode (mode_);

    return rv;
}


static gboolean
src_speech_set_modifiers_echo_type (SRCSpeechModifiersEchoType type)
{
    sru_assert (src_speech_initialized);

    src_speech_modifiers_echo_type = type;
    
    return TRUE;
}


static gboolean
src_speech_set_modifiers_echo_type_from_string (gchar *type)
{
    gboolean rv = TRUE;
    SRCSpeechModifiersEchoType type_ = src_speech_modifiers_echo_type;

    sru_assert (type);

    if (strcmp (type, "ALL") == 0)
	type_ = SRC_SPEECH_MODIFIERS_ECHO_ALL;
    else if (strcmp (type, "NONE") == 0)
	type_ = SRC_SPEECH_MODIFIERS_ECHO_NONE;
    else
    {
	sru_warning (_("value not supported for speech modifiers type \"%s\""), type);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_modifiers_echo_type (type_);

    return rv;
}


static gboolean
src_speech_set_cursors_echo_type (SRCSpeechCursorsEchoType type)
{
    sru_assert (src_speech_initialized);

    src_speech_cursors_echo_type = type;
    
    return TRUE;
}


static gboolean
src_speech_set_cursors_echo_type_from_string (gchar *type)
{
    gboolean rv = TRUE;
    SRCSpeechCursorsEchoType type_ = src_speech_cursors_echo_type;

    sru_assert (type);

    if (strcmp (type, "ALL") == 0)
	type_ = SRC_SPEECH_CURSORS_ECHO_ALL;
    else if (strcmp (type, "NONE") == 0)
	type_ = SRC_SPEECH_CURSORS_ECHO_NONE;
    else
    {
	sru_warning (_("value not supported for speech cursors type \"%s\""), type);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_cursors_echo_type (type_);

    return rv;
}


static gboolean
src_speech_set_spaces_echo_type (SRCSpeechSpacesEchoType type)
{
    sru_assert (src_speech_initialized);

    src_speech_spaces_echo_type = type;
    
    return TRUE;
}


static gboolean
src_speech_set_spaces_echo_type_from_string (gchar *type)
{
    gboolean rv = TRUE;
    SRCSpeechSpacesEchoType type_ = src_speech_spaces_echo_type;

    sru_assert (type);

    if (strcmp (type, "ALL") == 0)
	type_ = SRC_SPEECH_SPACES_ECHO_ALL;
    else if (strcmp (type, "NONE") == 0)
	type_ = SRC_SPEECH_SPACES_ECHO_NONE;
    else
    {
	sru_warning (_("value not supported for speech spaces type \"%s\""), type);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_spaces_echo_type (type_);

    return rv;
}


static gboolean
src_speech_set_dictionary_type (SRCSpeechDictionaryType type)
{
    sru_assert (src_speech_initialized);

    src_speech_dictionary_type = type;
    
    return TRUE;
}


static gboolean
src_speech_set_dictionary_type_from_string (gchar *type)
{
    gboolean rv = TRUE;
    SRCSpeechDictionaryType type_ = src_speech_dictionary_type;

    sru_assert (type);

    if (strcmp (type, "YES") == 0)
	type_ = SRC_SPEECH_DICTIONARY_YES;
    else if (strcmp (type, "NO") == 0)
	type_ = SRC_SPEECH_DICTIONARY_NO;
    else
    {
	sru_warning (_("value not supported for speech dictionary type \"%s\""), type);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_dictionary_type (type_);

    return rv;
}

static gboolean
src_speech_set_text_echo_type (SRCSpeechTextEchoType type)
{
    sru_assert (src_speech_initialized);
    
    src_speech_text_echo_type = type;
    
    return TRUE;
}


static gboolean
src_speech_set_text_echo_type_from_string (gchar *type)
{
    gboolean rv = TRUE;
    SRCSpeechTextEchoType type_ = src_speech_text_echo_type;

    sru_assert (type);

    if (strcmp (type, "WORD") == 0)
	type_ = SRC_SPEECH_TEXT_ECHO_WORD;
    else if (strcmp (type, "CHARACTER") == 0)
	type_ = SRC_SPEECH_TEXT_ECHO_CHAR;
    else
    {
	sru_warning (_("value not supported for speech text echo type \"%s\""), type);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_text_echo_type (type_);

    return rv;
}


static struct
{
    gchar  character;
    gchar *singular;
    gchar *plural;
    SRCSpeechPunctuationType type;
}src_speech_dictionary[]=
{
    {'~',	N_("Tilda"),		N_("Tildaes"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'!',	N_("Exclamation"),	N_("Exclamations"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'@',	N_("At"),		N_("Ats"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'#',	N_("Diez"),		N_("Diezes"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'$',	N_("Dollar"),		N_("Dollars"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'%',	N_("Percent"),		N_("Percents"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'^',	N_("Caret"),		N_("Carets"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'&',	N_("And"),		N_("Ands"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'*',	N_("Asterix"),		N_("Asterixes"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'(',	N_("Left Paranthesises"),N_("Left Paranthesises"),SRC_SPEECH_PUNCTUATION_SAME},
    {')',	N_("Right Paranthesises"),N_("Right Paranthesises"),SRC_SPEECH_PUNCTUATION_SAME},
    {'_',	N_("Underscore"),	N_("Underscores"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'+',	N_("Plus"),		N_("Pluses"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'=',	N_("Equal"),		N_("Equals"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'-',	N_("Minus"),		N_("Minuses"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'`',	N_("Quote"),		N_("Quotes"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'[',	N_("Left Bracket"),	N_("Left Brackets"),	SRC_SPEECH_PUNCTUATION_SAME},
    {']',	N_("Right Bracket"),	N_("Right Brackets"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'{',	N_("Left Brace"),	N_("Left Braces"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'}',	N_("Right Brace"),	N_("Right Braces"),	SRC_SPEECH_PUNCTUATION_SAME},
    {';',	N_("Semicolon"),	N_("Semicolons"),	SRC_SPEECH_PUNCTUATION_SAME},
    {':',	N_("Colon"),		N_("Colons"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'"',	N_("Double quote"),	N_("Double quotes"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'\'',	N_("Apostrophe"),	N_("Apostrophes"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'\\',	N_("Back Slash"),	N_("Back Slashes"),	SRC_SPEECH_PUNCTUATION_SAME},
    {'|',	N_("Bar"),		N_("Bars"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'>',	N_("Grant"),		N_("Grants"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'<',	N_("Less"),		N_("Lesses"),		SRC_SPEECH_PUNCTUATION_SAME},
    {',',	N_("Comma"),		N_("Commas"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'.',	N_("Dot"),		N_("Dots"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'/',	N_("Slash"),		N_("Slashes"),		SRC_SPEECH_PUNCTUATION_SAME},
    {'?',	N_("Introgation"),	N_("Introgations"),	SRC_SPEECH_PUNCTUATION_SAME}
};


static gboolean
src_speech_set_punctuation_type (SRCSpeechPunctuationType type)
{
    gint i;
    sru_assert (src_speech_initialized);

    for (i = 0; i < G_N_ELEMENTS (src_speech_dictionary); i++)
	src_speech_dictionary[i]. type = type;

    return TRUE;
}


static gboolean
src_speech_set_punctuation_type_from_string (gchar *type)
{
    gboolean rv = TRUE;
    SRCSpeechPunctuationType type_ = SRC_SPEECH_PUNCTUATION_SAME;

    sru_assert (type);

    if (strcmp (type, "IGNORE") == 0)
	type_ = SRC_SPEECH_PUNCTUATION_IGNORE;
    else if (strcmp (type, "SAME") == 0)
	type_ = SRC_SPEECH_PUNCTUATION_SAME;
    else if (strcmp (type, "TRANSLATE") == 0)
	type_ = SRC_SPEECH_PUNCTUATION_TRANSLATE;
    else if (strcmp (type, "BOTH") == 0)
	type_ = SRC_SPEECH_PUNCTUATION_BOTH;
    else
    {
	sru_warning (_("value not supported for speech punctuation type \"%s\""), type);
	rv = FALSE;
    }

    if (rv)
	rv = src_speech_set_punctuation_type (type_);

    return rv;
}


void
src_speech_shutup ()
{
    sru_assert (src_speech_initialized);

    src_speech_send ("<SRSOUT><SHUTUP/></SRSOUT>");
}


void
src_speech_pause ()
{
    sru_assert (src_speech_initialized);

    src_speech_send ("<SRSOUT><PAUSE/></SRSOUT>");
}


void
src_speech_resume ()
{
    sru_assert (src_speech_initialized);

    src_speech_send ("<SRSOUT><RESUME/></SRSOUT>");
}


gboolean
src_speech_send_chunk (gchar *chunk, SRCSpeechPriority priority)
{
    gchar *output;
    gboolean rv = FALSE;
    static SRCSpeechPriority last_priority = SRC_SPEECH_PRIORITY_IDLE;

    sru_assert (src_speech_initialized);
    sru_assert (chunk && chunk[0]);
    if (last_priority >= priority)
	src_speech_shutup ();
    output = g_strconcat ("<SRSOUT>", chunk, "</SRSOUT>", NULL);
    rv = src_speech_send (output);
    g_free (output);
    last_priority = priority;

    return rv;
}

static gboolean
src_dictionary_hash_init ()
{
    src_dictionary_hash = g_hash_table_new (g_str_hash , g_str_equal);
    if (!src_dictionary_hash)
	return FALSE;
    return TRUE;
}

static gboolean
src_dictionary_hash_terminate ()
{
    sru_assert (src_dictionary_hash != NULL);
    g_hash_table_destroy (src_dictionary_hash);
    src_dictionary_hash = NULL;
    return TRUE;
}

static gboolean
src_dictinary_add_item (gchar *key,
			gchar *value)
{
    sru_assert (src_dictionary_hash != NULL);
/*    fprintf (stderr,"%s-%s",key,value);*/
    g_hash_table_insert (src_dictionary_hash, 
			key,
			value);
    return TRUE;
}

/* DO NOT free the returned element*/
static gchar*
src_dictionary_lookup (gchar *key)
{
    gchar *word;

    sru_assert (key);
    sru_assert (src_dictionary_hash != NULL);
    word = g_hash_table_lookup (src_dictionary_hash, key);
    if (word)
	return word;
	
    return NULL;
}

static gboolean
src_dictionary_entry_split (const gchar *entry,
			    gchar **word,
			    gchar **replace)
{
    gchar *separator = NULL;
    
    *word = NULL;
    *replace = NULL;

    if (!entry)
	return FALSE;
    
    separator = g_utf8_strchr (entry, -1, '<');
    if (separator && *(g_utf8_find_next_char (separator, NULL)) == '>')
    {
	*word = g_strndup (entry, separator - entry);
	separator = g_utf8_find_next_char (separator, NULL);
	*replace  = g_strdup (g_utf8_find_next_char (separator,NULL));
	return TRUE;
    }
    return FALSE;
}

static gboolean
src_dictionary_load ()
{
    GSList *crt = NULL;
    GSList *dictionary_list = NULL;

    if (!srconf_get_data_with_default (SPEECH_DICTIONARY_LIST, CFGT_LIST,
				  &dictionary_list, NULL, 
				  SPEECH_DICTIONARY_SECTION))
    return FALSE;

    
    crt = dictionary_list;
    
    while (crt)
    {
	gchar *key   = NULL;
	gchar *value = NULL;

	src_dictionary_entry_split ((const gchar *)crt->data,
				    &key, &value);
		
	src_dictinary_add_item (key, value);
	
	crt = crt->next;
    }
  	        
    srconf_free_slist (dictionary_list);
    
    return TRUE;
}

static gboolean
src_dictionary_terminate ()
{
    src_dictionary_hash_terminate ();
    return TRUE;
}

static gboolean
src_dictionary_init ()
{

    if (!src_dictionary_hash_init ())
    {
	sru_warning (_("Unable to initialize hashtable"));
	return FALSE;
    }
    
    if (!src_dictionary_load ())
    {
	sru_warning (_("Unable to load dictionary from gconf"));
	return FALSE;
    }
    return TRUE;
}


static gboolean
src_speech_init_default_voice ()
{
    gboolean rv = FALSE;
    gchar *name, *volume, *rate, *pitch;
 
    sru_assert (src_speech_initialized);
    sru_assert (src_speech_engine_voices);
   
    name   = g_strdup (src_speech_engine_voices[0]);
    volume = g_strdup_printf ("%d", DEFAULT_SPEECH_VOLUME);
    rate   = g_strdup_printf ("%d", DEFAULT_SPEECH_RATE);
    pitch  = g_strdup_printf ("%d", DEFAULT_SPEECH_PITCH);
    rv = src_speech_set_voice_parameters ("default", name, volume, rate, pitch);

    g_free (name);
    g_free (rate);
    g_free (volume);
    g_free (pitch);	

    return rv;
}

static gboolean
src_speech_load_defaults ()
{
    gchar *tmp;

    sru_assert (src_speech_initialized);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_COUNT, CFGT_STRING, 
							    &tmp, NULL);
    if (tmp)
	src_speech_set_count_mode_from_string (tmp);
    else
        src_speech_set_count_mode (SRC_SPEECH_COUNT_NONE);
    g_free (tmp);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_PUNCTUATION, CFGT_STRING, 
							    &tmp, NULL);
    if (tmp)
	src_speech_set_punctuation_type_from_string (tmp);
    else
        src_speech_set_punctuation_type (SRC_SPEECH_PUNCTUATION_SAME);
    g_free (tmp);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_TEXT_ECHO, CFGT_STRING, 
							    &tmp, NULL);
    if (tmp)
	src_speech_set_text_echo_type_from_string (tmp);
    else
        src_speech_set_text_echo_type (SRC_SPEECH_TEXT_ECHO_CHAR);
    g_free (tmp);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_MODIFIERS, CFGT_STRING, 
							    &tmp, NULL);
    if (tmp)
	src_speech_set_modifiers_echo_type_from_string (tmp);
    else
        src_speech_set_modifiers_echo_type (SRC_SPEECH_MODIFIERS_ECHO_ALL);
    g_free (tmp);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_CURSORS, CFGT_STRING, 
							    &tmp, NULL);
    if (tmp)
	src_speech_set_cursors_echo_type_from_string (tmp);
    else
        src_speech_set_cursors_echo_type (SRC_SPEECH_CURSORS_ECHO_ALL);
    g_free (tmp);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_SPACES, CFGT_STRING, 
							    &tmp, NULL);
    if (tmp)
	src_speech_set_spaces_echo_type_from_string (tmp);
    else
        src_speech_set_spaces_echo_type (SRC_SPEECH_SPACES_ECHO_ALL);
    g_free (tmp);

    tmp = NULL;
    GET_SPEECH_CONFIG_DATA_WITH_DEFAULT (SPEECH_DICTIONARY_ACTIVE, CFGT_STRING, 
								   &tmp, NULL);
    if (tmp)
	src_speech_set_dictionary_type_from_string (tmp);
    else
        src_speech_set_dictionary_type (SRC_SPEECH_DICTIONARY_NO);
    g_free (tmp);

    return TRUE;
}


gboolean
src_speech_init ()
{
    sru_assert (!src_speech_initialized);
    
    src_speech_initialized = srs_xml_init (NULL);

    if (src_speech_initialized)
    {
	src_speech_engine_voices = src_speech_get_engine_voices ();
	src_speech_voices = src_speech_get_voices ();
    }

    if (src_speech_initialized && src_speech_engine_voices && src_speech_voices)
    {
	src_speech_initialized = src_speech_init_default_voice () &&
						    src_speech_voices_init ();
    }
    else
    {
	src_speech_initialized = FALSE;
    }

    if (src_speech_initialized)
	src_dictionary_init ();

    if (src_speech_initialized)
    {
	src_speech_set_spelling_mode (SRC_SPEECH_SPELL_NORMAL); /*FIXME: read from gconf????*/
	src_speech_load_defaults ();
    }
    srconf_unset_key (SRCORE_SPEECH_SENSITIVE, SRCORE_PATH);/*FIXME:why this line */
    SET_SRCORE_CONFIG_DATA (SRCORE_SPEECH_SENSITIVE, CFGT_BOOL, 
				&src_speech_initialized);/*FIXME: must be in speech gconf directory */

    if (src_speech_initialized)
    {
    	sru_message (_("speech initialization succeded"));
    }
    else
    {
    	sru_message (_("speech initialization failed"));
    }
    if (!src_speech_initialized)
    {
	g_strfreev (src_speech_voices);
	g_strfreev (src_speech_engine_voices);
    }

    
    return src_speech_initialized;
}


void
src_speech_terminate ()
{
    sru_assert (src_speech_initialized);

    srs_xml_terminate();
    src_speech_voices_terminate ();
    src_dictionary_terminate ();
    src_speech_initialized = FALSE;    
    g_strfreev (src_speech_voices);
    g_strfreev (src_speech_engine_voices);
}

#define src_speech_is_separator(x) (( x == '.' || x == ';' || x == '?' || x == ',' || x == '!' ) ? TRUE : FALSE)
static gchar*
src_dictionary_process_string (gchar *str)
{
    gchar *crt = str;
    gchar *word_start;
    gchar *rv = NULL;    
    gchar *word_key = NULL;
    gchar *word= NULL;
    gchar *tmp = NULL;

    
    g_assert (str != NULL);

    if (!str)
	return g_strdup (str);
	
    word_start = crt;
    
    while (*crt)
    {
	if (g_unichar_isspace (*crt) ||
	    g_unichar_ispunct (*crt))
	{
	    gchar *word_key = NULL;
	    gchar *word= NULL;
	    gchar *tmp = NULL;
	    
	    word_key = g_strndup (word_start, crt - word_start);

	    if ((word = src_dictionary_lookup (word_key)) == NULL)
		word = word_key;
	
	    tmp = g_strdup_printf ("%s%s%c", STR(rv), word, *crt);		
	    g_free (word_key);
	    g_free (rv);
	    rv = tmp;
	    word_start = g_utf8_next_char (crt);
	}

	crt = g_utf8_next_char (crt);
    }
    
    word_key = g_strndup (word_start, crt - word_start);
    
    if ((word = src_dictionary_lookup (word_key)) == NULL)
	word = word_key;
	
    tmp = g_strdup_printf ("%s%s%c", STR(rv), word, *crt);
		
    g_free (word_key);
    g_free (rv);
    rv = tmp;

    if (!rv)
	return g_strdup (str);
	        
    return rv;
}


static gint
src_speech_get_index_for_symbol (gchar c)
{
    gint i;
    
    sru_assert (src_speech_initialized);
    
    for (i = 0 ; i < G_N_ELEMENTS (src_speech_dictionary); i++)
    {
	if (src_speech_dictionary [i].character == c)
	{
	    return i;
	}
    }
    
    return -1;
}


static gchar*
src_speech_translate_symbol (gint index)
{
    gchar *rv = NULL;

    sru_assert (src_speech_initialized);
    sru_assert (index >= 0 && index < G_N_ELEMENTS (src_speech_dictionary));
    
    switch (src_speech_dictionary[index].type)
    {
	case SRC_SPEECH_PUNCTUATION_IGNORE:
	    break;
	case SRC_SPEECH_PUNCTUATION_SAME:
	    rv = g_strdup (" ");
	    rv[0] = src_speech_dictionary [index].character;
	    break;
	case SRC_SPEECH_PUNCTUATION_TRANSLATE:
	    rv = g_strdup (src_speech_dictionary[index].singular);
	    break;
	case SRC_SPEECH_PUNCTUATION_BOTH:
	    {
		gchar *tmp = g_strdup (" ");
		tmp[0] = src_speech_dictionary [index].character;
		rv = g_strconcat (tmp, " ", 
					src_speech_dictionary[index].singular,
					NULL);
		g_free (tmp);
	    }
	    break;
	default:
	    sru_assert_not_reached ();
	    break;
    }
    return rv;
}		


static gchar*
src_speech_count_symbol (gchar *start,
			 gchar **new_pos)
{
    gchar *crt = start + 1;
    gchar count[10];
    gchar *rv = NULL;
    gint cnt, index;

    sru_assert (src_speech_initialized);
    sru_assert (start && new_pos);
    index = src_speech_get_index_for_symbol (start [0]);
    sru_assert (index >= 0);

    while (*start == *crt)
	crt++;
    count [0] = '\0';
    cnt = crt- start;
    if (cnt > 1)
	sprintf (count, " %d ", crt - start);
    if (cnt == 1 && src_speech_is_separator (start[0]))
	rv = g_strndup (start, 1);
    else
	rv = g_strconcat (count, " ",
	    cnt == 1 ? src_speech_dictionary[index].singular : 
					src_speech_dictionary[index].plural,
	    " ", NULL);
    *new_pos = crt;
    return rv;
}
			 

extern gboolean src_enable_format_string; /*FIXME: must be removed */
gchar*
src_speech_process_string (const gchar   *text,
			   SRCSpeechCountMode cmode,
			   SRCSpeechSpellingMode smode)
{
    gchar *crt = (gchar*) text;
    gchar *rv;
    gchar *tmp;
    
    sru_assert (src_speech_initialized);
    sru_assert (text);

    if (cmode == SRC_SPEECH_COUNT_AUTO)
	cmode = src_speech_count_mode;
    if (smode == SRC_SPEECH_SPELL_AUTO)
	smode = src_speech_spelling_mode;

    if (!src_enable_format_string)
	return g_strdup (text);
	
    if (text[0] == ' ' && text[1] == '\0' && smode == SRC_SPEECH_SPELL_NORMAL)
	return g_strdup (_("space"));
    if (text[0] && text[1] == '\0')
	return g_strdup (text);

    if (src_speech_dictionary_type == SRC_SPEECH_DICTIONARY_YES)
    {
	crt = src_dictionary_process_string (crt);
	tmp = crt;
    }
    
    rv = g_strdup ("");
    while (*crt)
    {
	gchar *next = g_utf8_next_char (crt);
	gchar *add = NULL;

	if (next - crt == 1)
	{
	    gint index = src_speech_get_index_for_symbol (crt[0]);

	    if (index != -1)
	    {
		if (smode == SRC_SPEECH_SPELL_NORMAL && cmode == SRC_SPEECH_COUNT_NONE)
		    add = src_speech_translate_symbol (index);
		if (cmode == SRC_SPEECH_COUNT_ALL)
		    add = src_speech_count_symbol (crt, &next);
	    }
	    else
	    {
		add = g_strndup (crt, next - crt);
	    }
	}
	else
	{
	    add = g_strndup (crt, next - crt);
	}
	if (add)
	{
	    gchar *tmp = rv;
	    rv = g_strconcat (rv, add, NULL);
	    g_free (tmp);
	    g_free (add); 
	}
	crt = next;
    }
    
    
    if (src_speech_dictionary_type == SRC_SPEECH_DICTIONARY_YES)
	g_free (tmp);
    	
    return rv[0] ? rv : NULL;
}


gchar*
src_speech_get_text_voice_attributes (gchar *voice,
				      SRCSpeechSpellingMode smode)
{
    SRCSpeechSpellingMode mode = smode;
    gchar *spelling;

    sru_assert (src_speech_initialized);
    sru_assert (voice);
    
    if (smode == SRC_SPEECH_SPELL_AUTO)
	mode = src_speech_spelling_mode;

    if (mode == SRC_SPEECH_SPELL_MILITARY)
	spelling = " spelling='military'";
    else if (mode == SRC_SPEECH_SPELL_CHAR_BY_CHAR)
	spelling = " spelling='char'";
    else
	spelling = "";

    if (!src_speech_valid_voice (voice))
    {
	sru_warning (_("can not find \"%s\" voice. use default voice"), voice);
	voice = "default";
    }

    return g_strconcat ("voice=\"", voice, "\"", spelling, NULL);
}


#define SRC_SPEECH_PITCH_STEP 5
gboolean
src_speech_modify_pitch (SRCModifyMode mode)
{
    gchar ** voices;
    gchar *message;
    
    sru_assert (src_speech_initialized);
    
    if (mode == SRC_MODIF_DEFAULT)
	message = _("voices pitch will be set to default value");
    else if (mode == SRC_MODIF_DECREASE)
	message = _("voices pitch will be decreased");
    else
	message = _("voices pitch will be increased");
	
    src_speech_say_message (message);

    voices = src_speech_voices;
    if (voices)
    {
	gint i;
	for (i = 0; voices[i]; i++)
	{
	    gchar *pitch_ = NULL;
	    gint pitch;
	
	    if (mode != SRC_MODIF_DEFAULT)
	    {
		if (src_speech_get_voice_parameters (voices[i], NULL, NULL, NULL, &pitch_))
		{
		    pitch = atoi (pitch_);
		    g_free (pitch_);
		    pitch += (mode == SRC_MODIF_DECREASE) ? 
			    -SRC_SPEECH_PITCH_STEP : SRC_SPEECH_PITCH_STEP;
		}	
	    }
	    else
		pitch = DEFAULT_SPEECH_PITCH;
	    pitch = MIN (MAX (pitch, MIN_SPEECH_PITCH), MAX_SPEECH_PITCH);
	    pitch_ = g_strdup_printf ("%d", pitch);
	    if (src_speech_set_voice_parameters (voices[i], NULL, NULL, NULL, pitch_))
		src_speech_create_voice (voices[i], NULL, NULL, NULL, pitch_);
	    g_free (pitch_);	    
	}
    }
    return TRUE;
}


#define SRC_SPEECH_RATE_STEP 5
gboolean
src_speech_modify_rate (SRCModifyMode mode)
{
    gchar ** voices;
    gchar *message;
    
    sru_assert (src_speech_initialized);
    
    if (mode == SRC_MODIF_DEFAULT)
	message = _("voices rate will be set to default value");
    else if (mode == SRC_MODIF_DECREASE)
	message = _("voices rate will be decreased");
    else
	message = _("voices rate will be increased");
	
    src_speech_say_message (message);

    voices = src_speech_voices;
    if (voices)
    {
	gint i;
	for (i = 0; voices[i]; i++)
	{
	    gchar *rate_ = NULL;
	    gint rate;
	
	    if (mode != SRC_MODIF_DEFAULT)
	    {
		if (src_speech_get_voice_parameters (voices[i], NULL, NULL, &rate_, NULL))
		{
		    rate = atoi (rate_);
		    g_free (rate_);
		    rate += (mode == SRC_MODIF_DECREASE) ? 
			    -SRC_SPEECH_RATE_STEP : SRC_SPEECH_RATE_STEP;
		}	
	    }
	    else
		rate = DEFAULT_SPEECH_RATE;
	    rate = MIN (MAX (rate, MIN_SPEECH_RATE), MAX_SPEECH_RATE);
	    rate_ = g_strdup_printf ("%d", rate);
	    if (src_speech_set_voice_parameters (voices[i], NULL, NULL, rate_, NULL))
		src_speech_create_voice (voices[i], NULL, NULL, rate_, NULL);
	    g_free (rate_);	    
	}
    }

    return TRUE;
}


#define SRC_SPEECH_VOLUME_STEP 5
gboolean
src_speech_modify_volume (SRCModifyMode mode)
{
    gchar ** voices;
    gchar *message;
    
    sru_assert (src_speech_initialized);
    
    if (mode == SRC_MODIF_DEFAULT)
	message = _("voices volume will be set to default value");
    else if (mode == SRC_MODIF_DECREASE)
	message = _("voices volume will be decreased");
    else
	message = _("voices volume will be increased");
	
    src_speech_say_message (message);

    voices = src_speech_voices;
    if (voices)
    {
	gint i;
	for (i = 0; voices[i]; i++)
	{
	    gchar *volume_ = NULL;
	    gint volume;
	
	    if (mode != SRC_MODIF_DEFAULT)
	    {
		if (src_speech_get_voice_parameters (voices[i], NULL, &volume_, NULL, NULL))
		{
		    volume = atoi (volume_);
		    g_free (volume_);
		    volume += (mode == SRC_MODIF_DECREASE) ? 
			    -SRC_SPEECH_VOLUME_STEP : SRC_SPEECH_VOLUME_STEP;
		}
	    }
	    else
		volume = DEFAULT_SPEECH_VOLUME;
	    volume = MIN (MAX (volume, MIN_SPEECH_VOLUME), MAX_SPEECH_VOLUME);
	    volume_ = g_strdup_printf ("%d", volume);
	    if (src_speech_set_voice_parameters (voices[i], NULL, volume_, NULL, NULL))
		src_speech_create_voice (voices[i], NULL, volume_, NULL, NULL);
	    g_free (volume_);	    
	}
    }

    return TRUE;
}


gboolean
src_speech_proces_config_changed (SRConfigStructure *config)
{
    gboolean rv = TRUE;
    sru_assert (src_speech_initialized);
    sru_assert (config && config->key);
    
    if (strcmp (config->key, SPEECH_COUNT) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_count_mode_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_PUNCTUATION) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_punctuation_type_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_VOICE_TEST) == 0)
    {
    	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_test_voice (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_VOICE_CHANGES) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_add_voice (config->newvalue) &&
		    src_speech_init_voice (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_VOICE_REMOVED) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_remove_voice (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_TEXT_ECHO) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_text_echo_type_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_MODIFIERS) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_modifiers_echo_type_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_SPACES) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_spaces_echo_type_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_CURSORS) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_cursors_echo_type_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_DICTIONARY_ACTIVE) == 0)
    {
	sru_assert (config->type == CFGT_STRING);
	sru_assert (config->newvalue);
	rv = src_speech_set_dictionary_type_from_string (config->newvalue);
    }
    else if (strcmp (config->key, SPEECH_DICTIONARY_CHANGES) == 0)
    {
	sru_assert (config->type == CFGT_BOOL);
	sru_assert (config->newvalue);
	if (src_dictionary_terminate ())
	    rv = src_dictionary_init ();
    }
    else
    {
	sru_warning (_("no support for speech key \"%s\""), config->key);
	rv = FALSE;
    }

    return rv;
}



/*FIXME: Next functions must be removed */
gchar*
src_speech_get_voice(gchar *voice)
{
    return src_speech_get_text_voice_attributes (voice, SRC_SPEECH_SPELL_AUTO);
}

gboolean 
src_speech_say (gchar *message, gboolean shutup)
{
    return src_speech_say_message (message);
}

gchar*
src_process_string (const gchar   *text,
		    SRCSpeechCountMode  cmode,
		    SRCSpeechSpellingMode smode)
{
    if (!src_speech_initialized)
	return g_strdup (text);
    return src_speech_process_string (text, cmode, smode);
}

/*FIXME end*/
